"""
Forms for collecting and revoking signoffs.

Challenges:
 - the `Signet` form itself is just a labelled checkbox - it doesn't display any model fields.
   Yet, it should behave like a model form, defining a signoff instance once validated.
 - the form is only displayed when there is no instance - you can't edit a saved signoff, only revoke it.
   So these forms don't have an instance - they are only used to "add" a signoff.
 - the form itself likely needs an association to the Signoff Type so it can be rendered correctly.
 - so, the signoff type itself might be obtained from the request data (e.g., in a generic view)
   So if the request is bad, it may not be possible to construct a Form class to validate rest of data.

Signets with relations:
The form is being used to sign off on something specific, so it likely needs a relation to some concrete object.
So the view will need some way to know which object is being signed off on.

To solve this for concrete `Signets` with relational fields, try ONE of these approaches:
  - pre-create the signoff instance, with relations in place, and pass it to the form.
    The `Signet` instance, with its relations, will be saved if form validates and is signed.
  - specialize `AbstractSignoffForm` to add hidden fields with the extra relation data;
    Use initial data to populate these fields and
    be sure to override `clean()` to validate the extra data.
"""
from typing import Callable, Type, Union

from django import forms
from django.core.exceptions import ImproperlyConfigured, ValidationError

from signoffs.core.utils import class_service

opt_callable = Union[Type, Callable]


class AbstractSignoffForm(forms.ModelForm):
    """Abstract Base class for the signoff_form_factory"""

    signed_off = forms.BooleanField(label="I consent", required=False)
    signoff_id = forms.Field(widget=forms.HiddenInput, required=True)

    class Meta:
        model = None  # Concrete Forms must supply the signoff's signetModel: signoff.get_signetModel()
        exclude = ["user", "sigil", "sigil_label", "timestamp"]

    def __init__(self, *args, instance=None, **kwargs):
        """
        Form accepts an optional signoff, used like the instance parameter for ModelForms to pass initial values.
        Form also accepts 'user' as optional kwarg: the user who is signing off
        """
        if instance and instance.is_signed():
            raise ValueError(f"Attempt to edit a signed signoff! {instance}")
        self.signoff_instance = instance
        super().__init__(
            *args, instance=instance.signet if instance else None, **kwargs
        )

    def is_signed_off(self):
        """return True iff this form is signed off"""
        return self.is_valid() and self.cleaned_data.get("signed_off")

    def clean(self):
        """
        Validate signoff for consistency with the instance form was intialized with.

        :::{note}
        Don't be tempted to check permissions here!  The form is clean even if user doesn't have permission!
        :::
        """
        cleaned_data = super().clean()

        # the signoff returned from cleaned_data must match the form's signoff instance
        id = cleaned_data.get("signoff_id")
        if self.signoff_instance is not None and self.signoff_instance.id != id:
            raise ValidationError(
                f"Invalid signoff form - signoff type {type(self.signoff_instance)} does not match form {id}"
            )

        return cleaned_data

    def sign(self, user, commit=True):
        """
        Sign and save this form for the given user, without checking permissions (no business logic invoked!)

        :return: the saved signoff instance, or None if the signoff was not actually signed.

        :::{note}
        If signoff has m2m relations and commit==False, caller is responsible to call self.save_m2m()
        :::
        """
        signet = super().save(commit=False)
        if self.is_signed_off() and signet:
            signoff = signet.signoff
            signoff.sign(user=user, commit=commit)
            return signoff
        # nothing to do if it's not actually signed...

    def save(self, *args, **kwargs):
        """Disable normal form save() method - signoff forms must be signed by a user"""
        raise ImproperlyConfigured("Use the sign() method to save signoff forms!")


def signoff_form_factory(
    signoff_type,
    baseForm=AbstractSignoffForm,
    form_prefix=None,
    signoff_field_kwargs=None,
):
    """
    Returns a Form class suited to collecting a signoff.
    Not unlike modelform_factory, except the model is provided by the signoff_type.
    baseForm can be overridden to add additional fields and/or fully customize validation and save logic.
    Validation ensures the type of signoff in POST matches the type provided.
    Signing this form with a User performs a permissions check and saves the signoff, if required.
    """
    signoff_field_kwargs = signoff_field_kwargs or {}
    signoff_field_kwargs.setdefault("label", signoff_type.label)
    signoff_field_kwargs.setdefault("required", False)

    class SignoffForm(baseForm):
        class Meta(baseForm.Meta):
            model = signoff_type.get_signetModel()

        prefix = form_prefix

        signed_off = forms.BooleanField(**signoff_field_kwargs)

        signoff_id = forms.CharField(initial=signoff_type.id, widget=forms.HiddenInput)

    return SignoffForm


class AbstractSignoffRevokeForm(forms.Form):
    """
    Form used to validate requests to revoke a signoff - not really intended to be user-facing.
    Not intended to collect signoffs, but rather simply to house the validation logic for revoke requests,
        which may be delete rather than post requests.
    It is not a ModelForm so it remains as generic as possible, though in many ways a Signet ModelForm might be useful.
    """

    signoff_id = forms.CharField(widget=forms.HiddenInput)
    signet_pk = forms.IntegerField(widget=forms.HiddenInput)

    def __init__(self, *args, **kwargs):
        """
        Form requires the signetModel class of the signet to be revoked
        """
        super().__init__(*args, **kwargs)

    def _get_signoff_type(self):
        from signoffs import registry

        try:
            return registry.get_signoff_type(self.cleaned_data.get("signoff_id"))
        except ImproperlyConfigured as e:
            raise ValidationError(str(e))

    def _get_signet(self, signoff_type):
        signetModel = signoff_type.get_signetModel()
        try:
            return signetModel.objects.get(pk=self.cleaned_data.get("signet_pk"))
        except signetModel.DoesNotExist as e:
            raise ValidationError(str(e))

    def clean(self):
        """
        Validate the signoff type for the form's signet matches the form's signoff type

        :::{note}
        Don't be tempted to check permissions here!  The form is clean even if user doesn't have permission!
        :::
        """
        cleaned_data = super().clean()
        signoff_type = self._get_signoff_type()
        signet = self._get_signet(signoff_type) if signoff_type else None
        if not signoff_type or not signet or not signet.signoff_id == signoff_type.id:
            raise ValidationError(
                f"Invalid signoff type {signoff_type} does not match form signet {signet}"
            )

        signoff_instance = signet.signoff
        cleaned_data["signoff"] = signoff_instance
        return cleaned_data

    def revoke(self, user, commit=True):
        """
        Revoke the signoff validated by this form, and return the revoked signoff,
        without checking permissions (no business logic invoked!)

        If commit is False, the form is validated and the signoff to be revoked is returned - up to you to revoke it.
        """
        if not self.is_valid():
            raise ValueError(
                "Attempt to save an invalid form.  Always call is_valid() before saving!"
            )

        signoff = self.cleaned_data.get("signoff")
        if not user or not signoff:
            raise ValueError(
                f"Invalid User {user} or signoff {signoff} for revoke operation."
            )
        if commit:
            signoff.revoke(user)
        return signoff

    def save(self, *args, **kwargs):
        """Disable normal form save() method - signoff forms must be signed by a user"""
        raise ImproperlyConfigured("Use the revoke() method to revoke signoff forms!")


def revoke_form_factory(
    signoff_type,
    baseForm=AbstractSignoffRevokeForm,
    form_prefix=None,
    signoff_field_kwargs=None,
):
    """
    Returns a Form class suited to validation a signoff revoke request.
    Not unlike modelform_factory, except the model is provided by the signoff_type.
    baseForm can be overridden to add additional fields and/or fully customize validation and save logic.
    Validation ensures the type of signoff in POST matches the type provided.
    Revoking this form with a User performs a permissions check and deletes the signoff's signet, if required.
    """

    class SignoffRevokeForm(baseForm):
        prefix = form_prefix

        signoff_id = forms.CharField(initial=signoff_type.id, widget=forms.HiddenInput)

    return SignoffRevokeForm


class SignoffTypeForms:
    """Manage the forms used by a particular signoff type - usually injected using a FormsManager service"""

    signoff_form: opt_callable = AbstractSignoffForm
    """baseForm type passed to signoff_form_factory or a callable that returns a Form subclass"""

    revoke_form: opt_callable = AbstractSignoffRevokeForm
    """baseForm type passed to revoke_form_factory or a callable that returns a Form subclass"""

    def __init__(self, signoff_type, signoff_form=None, revoke_form=None):
        self.signoff_type = signoff_type
        self.signoff_form = (
            signoff_form or type(self).signoff_form
        )  # don't bind forms to instance!
        self.revoke_form = revoke_form or type(self).revoke_form

    # Forms for collecting & revoking signoffs

    @staticmethod
    def _get_form_class(candidate):
        """candidate may be a Form subclass or a callable that returns one - return the Form subclass either way"""
        return (
            candidate()
            if callable(candidate) and not isinstance(candidate, type)
            else candidate
        )

    def get_signoff_form_class(self, **kwargs):
        """Return a form class suitable for collecting a signoff of this Type.  kwargs passed through to factory."""
        form = self._get_form_class(self.signoff_form)
        kwargs.setdefault("baseForm", form)
        return signoff_form_factory(signoff_type=self.signoff_type, **kwargs)

    def get_revoke_form_class(self, **kwargs):
        """Return a form class suitable for validating revoke request for a signoff of this Type."""
        form = self._get_form_class(self.revoke_form)
        kwargs.setdefault("baseForm", form)
        return revoke_form_factory(signoff_type=self.signoff_type, **kwargs)

    def get_signoff_form(self, data=None, **kwargs):
        """Return a form instance suited to collecting this signoff type for simple case, no factory args required"""
        return self.get_signoff_form_class()(data=data, **kwargs)

    def get_revoke_form(self, data=None, **kwargs):
        """Return a form instance suited to revoking this signoff type for simple case, no factory args required"""
        return self.get_revoke_form_class()(data=data, **kwargs)


class SignoffFormsManager(class_service(service_class=SignoffTypeForms)):
    """
    A descriptor class that "injects" a `SignoffTypeForms` instance into a Signoff instance.

    To inject custom form services:
      - provide a custom service_class:  `forms=SignoffFormsManager(service_class=MyInstanceForms)`
      - OR specialize class attributes:
        `MySignoffFormsManager = utils.service(SignoffTypeForms, signoff_form=mySignoffForm)`
      - OR both... `MySignoffFormsManager = utils.service(MyInstanceForms)`
    """


__all__ = [
    "AbstractSignoffForm",
    "signoff_form_factory",
    "AbstractSignoffRevokeForm",
    "revoke_form_factory",
    "SignoffTypeForms",
    "SignoffFormsManager",
]



"""
    contrib.signets adds basic concrete Signet and RevokedSignet models
"""
from django.apps import AppConfig


class ContribSignetsConfig(AppConfig):
    name = "signoffs.contrib.signets"
    label = "signoffs_signets"
    default = True
    default_auto_field = "django.db.models.BigAutoField"



"""
App-independent tests for contrib models
"""
from django.test import TestCase

from signoffs import models, signoffs
from signoffs.core.tests import fixtures

simple = signoffs.SimpleSignoff.register(id="test.contrib.simple")
revokable = signoffs.RevokableSignoff.register(id="test.contrib.revokable")
irrevokable = signoffs.IrrevokableSignoff.register(id="test.contrib.irrevokable")


class ContribSignoffTypeTests(TestCase):
    def setUp(self):
        super().setUp()
        self.user = fixtures.get_user(
            username="anyone"
        )  # A user with no specific permissions

    def test_simple_signoff(self):
        signoff = simple()
        self.assertEqual(signoff.signet_model, models.Signet)
        self.assertIsNone(signoff.revoke_model)
        self.assertTrue(signoff.is_permitted_signer(self.user))
        self.assertTrue(signoff.is_permitted_revoker(self.user))
        signet = signoff.signet
        self.assertEqual(signet.signoff_type, type(signoff))
        self.assertEqual(signet.signoff, signoff)
        # save and revoke
        signoff = signoff.sign(user=self.user)
        self.assertTrue(signoff.can_revoke(self.user))

    def test_revokable_signoff(self):
        signoff = revokable()
        self.assertEqual(signoff.signet_model, models.Signet)
        self.assertEqual(signoff.revoke_model, models.RevokedSignet)
        self.assertTrue(signoff.is_permitted_signer(self.user))
        self.assertTrue(signoff.is_permitted_revoker(self.user))
        signet = signoff.signet
        self.assertEqual(signet.signoff_type, type(signoff))
        self.assertEqual(signet.signoff, signoff)
        # save and revoke
        signoff = signoff.sign(user=self.user)
        self.assertTrue(signoff.can_revoke(self.user))

    def test_irrevokable_signoff(self):
        signoff = irrevokable()
        self.assertEqual(signoff.signet_model, models.Signet)
        self.assertIsNone(signoff.revoke_model)
        self.assertTrue(signoff.is_permitted_signer(self.user))
        self.assertFalse(signoff.is_permitted_revoker(self.user))
        signet = signoff.signet
        self.assertEqual(signet.signoff_type, type(signoff))
        self.assertEqual(signet.signoff, signoff)
        # save and revoke
        signoff = signoff.sign(user=self.user)
        self.assertFalse(signoff.can_revoke(self.user))



"""
    Proxy for core.models to simplify import statements and hide core package structure from client code.

    isort:skip_file
"""
from django.apps import apps

from signoffs.core.models import (
    AbstractApprovalSignet,
    AbstractApprovalStamp,
    AbstractRevokedSignet,
    AbstractSignet,
)
from signoffs.core.models.fields import (
    ApprovalField,
    ApprovalSignoffSet,
    ApprovalSignoffSingle,
)
from signoffs.core.models.fields import RelatedApprovalDescriptor as RelatedApproval
from signoffs.core.models.fields import RelatedSignoffDescriptor as RelatedSignoff
from signoffs.core.models.fields import SignoffField, SignoffSet, SignoffSingle

if apps.is_installed("signoffs.contrib.signets"):
    from signoffs.contrib.signets.models import RevokedSignet, Signet

if apps.is_installed("signoffs.contrib.approvals"):
    from signoffs.contrib.approvals.models import RevokedSignet as RevokedApprovalSignet
    from signoffs.contrib.approvals.models import Signet as ApprovalSignet
    from signoffs.contrib.approvals.models import Stamp



from .approvals import ApprovalInstanceRenderer, ApprovalRenderer
from .process import ApprovalProcessInstanceRenderer, ApprovalProcessRenderer
from .signoffs import SignoffInstanceRenderer, SignoffRenderer



"""
    Convenience methods for common tasks
"""
from django.http import Http404
from django.shortcuts import get_object_or_404

from signoffs import registry


def get_signet_or_404(signoff_type, signet_pk):
    """Return Signet with given pk, for the given Signoff Type or id, or raise Http404"""
    signoff = registry.get_signoff_type(signoff_type)
    if signoff is None:
        raise Http404(f"No registered signoff with id: {signoff_type}")
    return get_object_or_404(
        signoff.get_signetModel(), pk=signet_pk, signoff_id=signoff.id
    )


def get_signoff_or_404(signoff_type, signet_pk):
    """Return Signoff of given type or id, backed by Signet with the given pk, or raise Http404"""
    signet = get_signet_or_404(signoff_type, signet_pk)
    return signet.signoff


def get_approval_stamp_or_404(approval_type, stamp_pk):
    """Return ApprovalStamp instance with given pk for the given Approval Type or id, or raise Http404"""
    approval = registry.get_approval_type(approval_type)
    if approval is None:
        raise Http404(f"No registered approval with id: {approval_type}")
    return get_object_or_404(
        approval.get_stampModel(), pk=stamp_pk, approval_id=approval.id
    )


def get_approval_or_404(approval_type, stamp_pk):
    """Return Approval of given type or id, backed by StampModel with given pk, or raise Http404"""
    stamp = get_approval_stamp_or_404(approval_type, stamp_pk)
    return stamp.approval



"""
It can be very convenient for a signoff to know the endpoints to POST / DELETE their `Signet`.
In fact, the default renderers assume a signoff knows the URLs for sign and revoke requests.

This component can be extended to provide flexible url services to `Signoff` instances.
They are generally "injected" into the Signoff Type using a `SignoffUrlsManager` service descriptor
"""

from django.urls import reverse

from signoffs.core.utils import service


class SignoffInstanceUrls:
    """
    Defines the urls for Signing and Revoking a Signoff instance
    """

    # Define URL patterns for saving and revoking signoffs
    save_url_name: str = ""
    revoke_url_name: str = ""

    def __init__(self, signoff_instance, save_url_name=None, revoke_url_name=None):
        """Override default actions, or leave parameter None to use class default"""
        self.signoff = signoff_instance
        self.save_url_name = save_url_name or self.save_url_name
        self.revoke_url_name = revoke_url_name or self.revoke_url_name

    def get_save_url(self, args=None, kwargs=None):
        """Return the URL for requests to save the signoff"""
        args = args or ()
        kwargs = kwargs or {}
        return (
            reverse(self.save_url_name, args=args, kwargs=kwargs)
            if self.save_url_name
            else ""
        )

    def get_revoke_url(self, args=None, kwargs=None):
        """Return the URL for requests to revoke this signoff"""
        args = args or (self.signoff.signet.pk,)
        kwargs = kwargs or {}
        return (
            reverse(self.revoke_url_name, args=args, kwargs=kwargs)
            if self.revoke_url_name
            else ""
        )


class SignoffUrlsManager(service(SignoffInstanceUrls)):
    """
    A descriptor class that "injects" a SignoffInstanceUrls instance into a Signoff instance.

    To inject custom URL services:
      - provide a custom service_class:  `urls=SignoffUrlsManager(service_class=MyInstanceUrls)`
      - OR specialize class attributes:
        `SignoffUrlsManager = utils.service(SignoffInstanceUrls, revoke_url_name="my:app:revoke_view")`
      - OR both... `SignoffUrlsManager = utils.service(MyInstanceUrls)`

    """



"""
    Objects that know how to assess the status of an Approval
"""
from signoffs.core import utils


class ApprovalInstanceStatus:
    """Access status info about an Approval instance"""

    def __init__(self, approval_instance, **kwargs):
        """A status instance for given approval instance"""
        self.approval = approval_instance

    # Note: there is an intermediate state in which a pre-condition to approval is not yet met,
    #       so the approval is not yet "awaiting".  Controlled at approval-process level, not available on approval.
    def get_status(self):
        """Return a short status indicator, e.g., for use as a CSS class, dictionary key, etc."""
        return "complete" if self.approval.is_approved() else "awaiting"

    def get_label(self):
        """Return a string with a pithy label indicating the status of the approval."""
        return (
            "Approval Complete" if self.approval.is_approved() else "Awaiting Approval"
        )

    def get_css_class(self):
        """Return a CSS class used, e.g., to style the approval label in templates"""
        return "success" if self.approval.is_approved() else "warning"


class ApprovalStatus(utils.service(ApprovalInstanceStatus)):
    """
    A descriptor that "injects" a ApprovalInstanceStatus instance into a Approval instance.

    To inject custom rendering services:
      - provide a custom service_class:  `status=ApprovalStatus(service_class=MyInstanceStatus)`
      - OR specialize class attributes:
        `MyApprovalStatus = utils.service(ApprovalInstanceStatus, signet_template='my.tmpl.html')`
      - OR both... `MyApprovalStatus = utils.service(MyInstanceStatus)`

    """



"""
App-independent tests for view.actions - no app logic
"""
import django_fsm as fsm
from django.db.models import Model, TextChoices
from django.test import TestCase

import signoffs.core.signing_order as so
from signoffs.core import process as signoffs_process
from signoffs.core.approvals import BaseApproval
from signoffs.registry import register
from signoffs.signoffs import SignoffLogic
from signoffs.views import actions

from ..models.fields import ApprovalField
from . import fixtures, models


class ActionsTestsSignoff(models.BasicSignoff):
    signetModel = models.Signet
    label = "Consent?"


signoff_type = ActionsTestsSignoff.register(
    id="test.actions_signoff", logic=SignoffLogic(perm="auth.add_signoff")
)


class ActionsSignoffFormTests(TestCase):
    def setUp(self):
        self.user = fixtures.get_user(perms=("add_signoff",))
        self.data = dict(
            signoff_id=signoff_type.id,
            signed_off="on",
        )

    def get_signoff_form(self, data=None):
        """Return a signoff form instance bound to class self.data"""
        data = data or self.data
        a = actions.BasicUserSignoffActions(self.user, data)
        return a.forms.get_signoff_form()

    def test_create(self):
        self.assertTrue(
            type(self.get_signoff_form()), signoff_type.forms.get_signoff_form_class()
        )

    def test_is_signed_off(self):
        bf = self.get_signoff_form()
        self.assertTrue(bf.is_valid())
        self.assertTrue(bf.is_signed_off())

    def test_not_signed_off(self):
        bf = self.get_signoff_form(data=dict(signoff_id=signoff_type.id))
        self.assertTrue(bf.is_valid())
        self.assertFalse(bf.is_signed_off())

    def test_invalid_signoff(self):
        bf = self.get_signoff_form(
            data=dict(signed_off="True", signoff_id="invalid.type")
        )
        self.assertIsNone(bf)

    def test_invalid_data(self):
        bf = self.get_signoff_form(data=dict(signed_off="True"))
        self.assertIsNone(bf)


class BasicUserSignoffActionsTests(TestCase):
    def setUp(self):
        self.user = fixtures.get_user(perms=("add_signoff",))
        self.data = dict(
            signoff_id=signoff_type.id,
            signed_off="on",
        )

    def test_create(self):
        action = actions.BasicUserSignoffActions(self.user, self.data)
        self.assertEqual(action.forms.get_signoff_type(), signoff_type)

    def test_is_valid_signoff_request(self):
        u = fixtures.get_user()
        u_action = actions.BasicUserSignoffActions(u, self.data)
        self.assertFalse(u_action.validator.is_valid_signoff_request(signoff_type()))

        action = actions.BasicUserSignoffActions(self.user, self.data)
        self.assertTrue(action.validator.is_valid_signoff_request(signoff_type()))

    def test_sign_signoff(self):
        action = actions.BasicUserSignoffActions(self.user, self.data)
        self.assertTrue(action.sign_signoff())
        self.assertTrue(action.signoff.is_signed())

    def test_sign_signoff_no_commit(self):
        action = actions.BasicUserSignoffActions(self.user, self.data)
        self.assertTrue(action.sign_signoff(commit=False))
        self.assertFalse(action.signoff.is_signed())

        action.signoff.save()
        self.assertTrue(action.signoff.is_signed())

    def test_sign_signoff_invalid(self):
        data = dict(signed_off="True", signoff_id="invalid.type")
        d_action = actions.BasicUserSignoffActions(self.user, data)
        self.assertFalse(d_action.sign_signoff())

    def test_unsigned_signoff(self):
        data = dict(signoff_id=signoff_type.id)
        d_action = actions.BasicUserSignoffActions(self.user, data)
        self.assertFalse(d_action.sign_signoff())

    def test_sign_signoff_permission_denied(self):
        u = fixtures.get_user()
        u_action = actions.BasicUserSignoffActions(u, self.data)
        self.assertFalse(u_action.sign_signoff())

    def revoke_data(self):
        """sign a signoff and return data defining its revocation"""
        action = actions.BasicUserSignoffActions(self.user, self.data)
        action.sign_signoff()
        self.assertTrue(action.signoff.is_signed())
        return dict(
            signoff_id=signoff_type.id,
            signet_pk=action.signoff.signet.pk,
        )

    def test_is_valid_revoke_request(self):
        action = actions.BasicUserSignoffActions(self.user, self.data)
        action.sign_signoff()

        u = fixtures.get_user()
        u_action = actions.BasicUserSignoffActions(u, self.revoke_data())
        self.assertFalse(u_action.validator.is_valid_revoke_request(action.signoff))

        r_action = actions.BasicUserSignoffActions(self.user, self.revoke_data())
        self.assertTrue(r_action.validator.is_valid_revoke_request(action.signoff))

    def test_invalid_revoke_request(self):
        action = actions.BasicUserSignoffActions(self.user, self.data)
        action.sign_signoff()

        r_action = actions.BasicUserSignoffActions(
            self.user,
            self.revoke_data(),
            validator=actions.BasicSignoffValidator(
                user=self.user, verify_signet=lambda signoff: False
            ),
        )

        self.assertFalse(r_action.validator.is_valid_revoke_request(action.signoff))

    def test_revoke_signoff(self):
        r_action = actions.BasicUserSignoffActions(self.user, self.revoke_data())
        self.assertTrue(r_action.revoke_signoff(commit=True))
        self.assertFalse(r_action.signoff.is_signed())

    def test_revoke_signoff_no_commit(self):
        r_action = actions.BasicUserSignoffActions(self.user, self.revoke_data())
        self.assertTrue(r_action.revoke_signoff(commit=False))
        self.assertTrue(r_action.signoff.is_signed())

        r_action.revoke_signoff()
        self.assertFalse(r_action.signoff.is_signed())

    def test_revoke_signoff_invalid(self):
        r_data = {**self.revoke_data(), "signet_pk": "invalid.type"}
        r_action = actions.BasicUserSignoffActions(self.user, r_data)
        self.assertFalse(r_action.revoke_signoff())

        r_data = {**self.revoke_data(), "signet_pk": 123}
        r_action = actions.BasicUserSignoffActions(self.user, r_data)
        self.assertFalse(r_action.revoke_signoff())

    def test_revoke_signoff_permission_denied(self):
        u = fixtures.get_user()
        u_action = actions.BasicUserSignoffActions(u, self.revoke_data())
        self.assertFalse(u_action.revoke_signoff())


@register(id="signoffs.test.actions_approval")
class ActionsApproval(BaseApproval):
    stampModel = models.Stamp
    label = "Test Approval"

    first_signoff = models.ApprovalSignoff.register(
        id="test.approval.actions.first", logic=SignoffLogic(perm="auth.add_signoff")
    )
    final_signoff = models.ApprovalSignoff.register(
        id="test.approval.actions.final", logic=SignoffLogic(perm="auth.add_signoff")
    )

    signing_order = so.SigningOrder(first_signoff, final_signoff)


class BasicUserApprovalActionsTests(TestCase):
    def setUp(self):
        self.user = fixtures.get_user(perms=("add_signoff",))
        self.approval = ActionsApproval.create()

    def get_data(self, signoff=None, stamp=None, signed_off="on"):
        signoff = signoff or self.approval.get_next_signoff(for_user=self.user)
        stamp = stamp or self.approval.stamp
        return dict(
            signoff_id=signoff.id if signoff else self.approval.first_signoff.id,
            signed_off=signed_off,
            stamp=stamp.id,
        )

    def test_create(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        self.assertEqual(type(action.signoff_actions), actions.BasicUserSignoffActions)
        self.assertEqual(
            action.signoff_actions.forms.get_signoff_type(), self.approval.first_signoff
        )

    def test_is_valid_approval_signoff_request(self):
        u = fixtures.get_user()
        u_action = actions.BasicUserApprovalActions(u, self.get_data(), self.approval)
        self.assertFalse(
            u_action.validator.is_valid_signoff_request(self.approval.first_signoff())
        )

        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        self.assertTrue(
            action.validator.is_valid_signoff_request(self.approval.first_signoff())
        )

    def test_sign_signoff(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        self.assertTrue(action.sign_signoff())
        self.assertTrue(action.signoff.is_signed())
        self.assertEqual(self.approval.signoffs.count(), 1)

    def test_sign_to_approval(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        self.assertFalse(self.approval.is_approved())

        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        self.assertTrue(self.approval.is_approved())

    def test_sign_to_approval_manual(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()

        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff(commit=False)
        self.assertFalse(self.approval.is_approved())
        action.sign_signoff()
        self.assertTrue(self.approval.is_approved())

    def revoke_data(self, signoff=None):
        """sign a signoff and return data defining its revokation"""
        if not signoff:
            action = actions.BasicUserApprovalActions(
                self.user, self.get_data(), self.approval
            )
            action.sign_signoff()
            self.assertTrue(action.signoff.is_signed())
            signoff = action.signoff
        return dict(
            signoff_id=signoff.id,
            signet_pk=signoff.signet.pk,
            stamp=signoff.signet.stamp_id,
        )

    def test_is_valid_revoke_signoff_request(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        signoff = action.signoff

        u = fixtures.get_user()
        u_action = actions.BasicUserApprovalActions(
            u, self.revoke_data(signoff=signoff), self.approval
        )
        self.assertFalse(u_action.validator.is_valid_revoke_request(signoff))

        r_action = actions.BasicUserApprovalActions(
            self.user, self.revoke_data(signoff=signoff), self.approval
        )
        self.assertTrue(r_action.validator.is_valid_revoke_request(signoff))

    def test_revoke_signoff(self):
        r_action = actions.BasicUserApprovalActions(
            self.user, self.revoke_data(), self.approval
        )
        self.assertTrue(r_action.revoke_signoff())
        self.assertFalse(r_action.signoff.is_signed())

    def test_revoke_signoff_manual(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        signoff = action.signoff

        r_action = actions.BasicUserApprovalActions(
            self.user, self.revoke_data(signoff=signoff), self.approval
        )
        self.assertTrue(r_action.revoke_signoff(commit=False))
        self.assertTrue(r_action.signoff.is_signed())
        r_action.revoke_signoff()
        self.assertFalse(r_action.signoff.is_signed())

    def test_approve_request(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        self.assertFalse(action.is_valid_approve_request())

        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        self.approval.stamp.approved = False
        self.assertTrue(action.is_valid_approve_request())
        action.approve()
        self.assertTrue(self.approval.is_approved())

    def test_revoke_approve_request(self):
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        self.assertFalse(action.is_valid_approval_revoke_request())
        action = actions.BasicUserApprovalActions(
            self.user, self.get_data(), self.approval
        )
        action.sign_signoff()
        self.assertTrue(self.approval.is_approved())

        self.assertTrue(action.is_valid_approval_revoke_request())
        self.assertTrue(action.revoke_approval())
        self.assertFalse(self.approval.is_approved())
        self.assertEqual(self.approval.signoffs.count(), 0)


class FsmActionsProcessModel(Model):
    """An Approval Process model where transitions and approval sequencing are defined by FSM"""

    class States(TextChoices):
        STATE0 = "Initiated"
        STATE1 = "State 1"
        STATE2 = "State 2"

    state = fsm.FSMField(choices=States.choices, default=States.STATE0)

    approval1, approval1_stamp = ApprovalField(ActionsApproval)
    approval2, approval2_stamp = ApprovalField(
        ActionsApproval.register("signoffs.test.actions_approval2")
    )

    process = signoffs_process.FsmApprovalsProcess()

    # FSM approval state transitions
    @process.approval_transition(
        approval1, state, source=States.STATE0, target=States.STATE1
    )
    def transition1(self, approval):
        pass

    @process.revoke_transition(
        approval1, state, source=States.STATE1, target=States.STATE0
    )
    def revoke1(self, approval):
        pass

    @process.approval_transition(
        approval2, state, source=States.STATE1, target=States.STATE2
    )
    def transition2(self, approval):
        pass

    @process.revoke_transition(
        approval2, state, source=States.STATE2, target=States.STATE1
    )
    def revoke2(self, approval):
        pass

    def sign_and_approve(self, user=None):
        """Helper for testing - attempt to complete signoffs and make next approval transition."""
        user = user or fixtures.get_user(perms=("add_signoff",))
        approval = self.process.get_next_available_approval()
        if not approval:
            return
        while not approval.ready_to_approve():
            signoff = approval.get_next_signoff(for_user=user)
            if not signoff:
                break
            signoff.sign_if_permitted(user)
        self.process.try_approve_transition(approval, user)


class ApprovalProcessUserActionsTests(TestCase):
    def setUp(self):
        self.user = fixtures.get_user(perms=("add_signoff",))
        self.fsm = FsmActionsProcessModel.objects.create()
        self.approval_process = self.fsm.process

    def get_data(self, signoff=None, stamp=None, signed_off="on"):
        approval = self.approval_process.get_next_approval()
        if not approval:
            return {}
        signoff = signoff or approval.get_next_signoff(for_user=self.user)
        stamp = stamp or approval.stamp
        return dict(
            signoff_id=signoff.id if signoff else approval.first_signoff.id,
            signed_off=signed_off,
            stamp=stamp.id,
        )

    def test_create(self):
        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        self.assertEqual(type(action.signoff_actions), actions.BasicUserSignoffActions)
        self.assertEqual(
            action.signoff_actions.forms.get_signoff_type(),
            self.fsm.approval1.first_signoff,
        )

    def test_sign_signoff(self):
        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        self.assertTrue(action.sign_signoff())
        self.assertTrue(action.signoff.is_signed())
        self.assertEqual(action.approval.signoffs.count(), 1)

    def test_sign_to_approval(self):
        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        action.sign_signoff()
        self.assertFalse(action.approval.is_approved())
        self.assertEqual(self.fsm.state, self.fsm.States.STATE0)

        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        action.sign_signoff()
        self.assertTrue(action.approval.is_approved())
        self.assertEqual(self.fsm.state, self.fsm.States.STATE1)

    def test_sign_to_approval2(self):
        self.fsm.sign_and_approve(self.user)
        self.assertEqual(self.fsm.state, self.fsm.States.STATE1)
        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        action.sign_signoff()
        self.assertFalse(action.approval.is_approved())
        self.assertEqual(self.fsm.state, self.fsm.States.STATE1)

        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        action.sign_signoff()
        self.assertTrue(action.approval.is_approved())
        self.assertEqual(self.fsm.state, self.fsm.States.STATE2)

    def revoke_data(self, signoff=None):
        """sign a signoff and return data defining its revokation"""
        if not signoff:
            approval = self.approval_process.get_next_approval()
            if not approval:
                return {}
            signoff = approval.signoffs[-1]
        return dict(
            signoff_id=signoff.id,
            signet_pk=signoff.signet.pk,
            stamp=signoff.signet.stamp_id,
        )

    def test_is_valid_approval_revoke_request(self):
        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        action.sign_signoff()
        r_action = actions.ApprovalProcessUserActions(
            self.user, {}, self.approval_process
        )
        self.assertFalse(r_action.is_valid_approval_revoke_request())

        self.fsm.sign_and_approve(self.user)
        self.assertEqual(self.fsm.state, self.fsm.States.STATE1)
        r_action = actions.ApprovalProcessUserActions(
            self.user,
            {},
            self.approval_process,
            approval=self.approval_process.get_approved_approvals()[-1],
        )
        self.assertTrue(r_action.is_valid_approval_revoke_request())

    def test_revoke_signoff(self):
        action = actions.ApprovalProcessUserActions(
            self.user, self.get_data(), self.approval_process
        )
        action.sign_signoff()
        r_action = actions.ApprovalProcessUserActions(
            self.user, self.revoke_data(action.signoff), self.approval_process
        )
        self.assertTrue(r_action.revoke_signoff())
        self.assertFalse(r_action.signoff.is_signed())

    def test_approve_request(self):
        self.fsm.sign_and_approve(self.user)
        approval = self.approval_process.get_approved_approvals()[-1]
        approval.stamp.approved = False
        self.fsm.state = self.fsm.States.STATE0
        action = actions.ApprovalProcessUserActions(
            self.user, {}, self.approval_process
        )
        action.approve()
        self.assertTrue(approval.is_approved())
        self.assertEqual(self.fsm.state, self.fsm.States.STATE1)

    def test_revoke_approve_request(self):
        self.fsm.sign_and_approve(self.user)
        self.fsm.sign_and_approve(self.user)
        self.assertEqual(self.fsm.state, self.fsm.States.STATE2)

        action = actions.ApprovalProcessUserActions(
            self.user,
            {},
            self.approval_process,
            approval=self.approval_process.get_approved_approvals()[-1],
        )
        self.assertTrue(action.is_valid_approval_revoke_request())
        self.assertTrue(action.revoke_approval())
        self.assertFalse(action.approval.is_approved())
        self.assertEqual(action.approval.signoffs.count(), 0)
        self.assertEqual(self.fsm.state, self.fsm.States.STATE1)

        action = actions.ApprovalProcessUserActions(
            self.user,
            {},
            self.approval_process,
            approval=self.approval_process.get_approved_approvals()[-1],
        )
        self.assertTrue(action.revoke_approval())
        self.assertEqual(self.fsm.state, self.fsm.States.STATE0)



"""
App-independent tests for Approval model descriptors - no app logic
"""
from django.contrib.auth.models import AnonymousUser
from django.test import TestCase

from . import fixtures
from .models import LeaveApproval, LeaveRequest, Stamp


class SimpleApprovalRelationTests(TestCase):
    def test_approvalfield_relations(self):
        lr = LeaveRequest()
        self.assertEqual(LeaveRequest.approval.id, LeaveApproval.id)
        self.assertEqual(type(lr.approval.stamp), LeaveApproval.get_stampModel())
        # ApprovalField descriptor creates stamp on first access to ensure subsequent signoff relations can be saved.
        self.assertIsNotNone(lr.approval_stamp)

    def test_approvalfield(self):
        lr = LeaveRequest()
        # OneToOne forward relation
        self.assertTrue(isinstance(lr.approval, LeaveApproval))
        self.assertEqual(lr.approval.stamp_model, Stamp)
        self.assertFalse(lr.approval.is_approved())


class ApprovalRelationTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.u1 = fixtures.get_user()
        cls.u2 = fixtures.get_user()
        cls.u3 = fixtures.get_user()
        lr = LeaveRequest.objects.create()
        lr.employee_signoff.sign(cls.u1)
        cls.hr_signoffs = (
            lr.hr_signoffs.create(user=cls.u1),
            lr.hr_signoffs.create(user=cls.u2),
        )
        lr.mngmt_signoffs.create(user=cls.u3),
        lr.approval.get_next_signoff(cls.u1).sign(
            cls.u1
        )  # touching approval field is enough to create the approval relation
        lr.approval.get_next_signoff(cls.u2).sign(cls.u2)
        lr.save()
        cls.lr = lr

    def test_approvalfield(self):
        with self.assertNumQueries(1):
            lr = LeaveRequest.objects.select_related("approval_stamp").get(
                pk=self.lr.pk
            )
            # OneToOne forward relation
            self.assertFalse(lr.approval.is_approved())

    def test_approval_signatories(self):
        with self.assertNumQueries(3):
            lr = (
                LeaveRequest.objects.select_related("approval_stamp")
                .prefetch_related("approval_stamp__signatories__user")
                .get(pk=self.lr.pk)
            )
            self.assertTrue(lr.approval.can_sign(self.u3))
            self.assertFalse(lr.approval.can_sign(AnonymousUser()))
            self.assertTrue(lr.approval.has_signed(self.u1))
            self.assertFalse(lr.approval.has_signed(self.u3))
            self.assertTrue(lr.approval.can_sign(self.u3))

    def test_approval_revoke(self):
        u = fixtures.get_user()
        lr = LeaveRequest.objects.select_related("approval_stamp").get(pk=self.lr.pk)
        lr.approval.approve()
        lr = LeaveRequest.objects.select_related("approval_stamp").get(pk=self.lr.pk)
        self.assertTrue(lr.approval.is_approved())
        lr.approval.revoke(u)
        lr = (
            LeaveRequest.objects.select_related("approval_stamp")
            .prefetch_related("approval_stamp__signatories")
            .get(pk=self.lr.pk)
        )
        self.assertFalse(lr.approval.is_approved())
        self.assertFalse(lr.approval.has_signatories())



"""
A `Signet` is a relation between a user and a timestamp.
A concrete `Signet` may have other information about, or perhaps a relation to, the thing being signed off.

A `Signet` records a one-time action that can generally only be completed by a `User` with permission.
The presence of a `Signet` record provides evidence that a particular someone signed off on a particular something.
A `Signet` is not intended to be edited.  Create them, revoke them, re-create them.  Don't edit and re-save them.
To revoke a `Signet`, we can simply delete the `Signet` record.
To maintain a "blame" history, we can instead record who and when the signet was revoked with a `RevokedSignet`.
"""
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import (
    FieldError,
    ImproperlyConfigured,
    PermissionDenied,
    ValidationError,
)
from django.db import models
from django.utils import timezone

from signoffs import settings


class SignetQuerySet(models.QuerySet):
    """
    Custom queries for signets
    """

    def active(self):
        """Filter out revoked signets"""
        try:
            return self.filter(revoked=None)
        except FieldError:  # caveat: not every signet model has a related revoke model.
            return self

    def revoked(self):
        """Return only revoked signets"""
        try:
            return self.exclude(revoked=None)
        except FieldError:  # no related manager  --> no revoked signets
            return self.none()

    def with_user(self):
        """Select the related signing User"""
        return self.select_related("user")

    def signoffs(self, signoff_id=None, subject=None):
        """
        Returns list of signoff objects, one for each signet in queryset,
            optionally filtered for specific signoff type - filtering done in-memory for performance.
        """
        return [
            signet.get_signoff(subject=subject)
            for signet in self
            if signoff_id is None or signet.signoff_id == signoff_id
        ]


BaseSignetManager = models.Manager.from_queryset(SignetQuerySet)


class ActiveSignetManager(BaseSignetManager):
    """Filters out revoked signets from queryset - should be the default manager"""

    def get_queryset(self):
        return super().get_queryset().active()


class RevokedSignetManager(BaseSignetManager):
    """Filters out un-revoked signets from queryset (only revoked signets returned"""

    def get_queryset(self):
        return super().get_queryset().revoked()


def validate_signoff_id(value):
    """Raise ValidationError if value is not a registered Signoff Type ID"""
    from signoffs import registry

    if value is None or value not in registry.signoffs:
        raise ValidationError(f"Invalid or unregistered signoff {value}")


def get_signet_defaults(signet):
    """
    Return a dictionary of default values for fields the given signet -
        this is the default implementation for settings.SIGNOFFS_SIGNET_DEFAULTS setting.
    Called just before signet is saved - signet is guaranteed to have a valid user object.
    Dictionary keys must match signet model field names - only editable fields of signet may have defaults.
    Only fields with no value will be set to its default value.
    """
    return {
        "sigil": signet.user.get_full_name() or signet.user.username,
        "sigil_label": signet._meta.get_field("sigil").verbose_name,
    }


class AbstractSignet(models.Model):
    """
    Abstract base class for all Signet models
    A Signet is the model for a single "signature" or a user's personal "seal" or "sigil"
    Persistence layer for a signoff: who, when, what (what is supplied by concrete class, e.g., with a FK to other model)
    Note: user relation is required for signing - enforced by app logic rather than at DB level so that...
    SET_NULL on_delete for user field is sensible for use-cases where signoff should persist even after user is deleted.
       For other on_delete behaviours, concrete RevokedSignet classes will need to override the user relation.
    """

    signoff_id = models.CharField(
        max_length=100,
        null=False,
        validators=[validate_signoff_id],
        verbose_name="Signoff Type",
    )
    user = models.ForeignKey(
        get_user_model(), on_delete=models.SET_NULL, null=True, related_name="+"
    )
    sigil = models.CharField(max_length=256, null=False, verbose_name="Signed By")
    sigil_label = models.CharField(
        max_length=256, null=True
    )  # optional label, e.g., signatory's title or role
    timestamp = models.DateTimeField(default=timezone.now, editable=False, null=False)

    class Meta:
        abstract = True
        ordering = [
            "timestamp",
        ]  # chronological order

    objects = ActiveSignetManager()  # default manager excludes revoked signets.
    revoked_signets = RevokedSignetManager()

    read_only_fields = (
        "signoff_id",
        "timestamp",
    )  # fields managed in code cannot be manipulated

    def __str__(self):
        return (
            f"{self.signoff_id} by {self.user} at {self.timestamp}"
            if self.is_signed()
            else f"{self.signoff_id} (unsigned)"
        )

    @property
    def signoff_type(self):
        """Return the Signoff Type (class) that governs this signet"""
        from signoffs.registry import signoffs

        signoff_type = signoffs.get(self.signoff_id)
        if not signoff_type:
            raise ImproperlyConfigured(
                f"""Signoff type {signoff_type} not registered.
            See AUTODISCOVER settings to discover signoff types when django loads."""
            )
        return signoff_type

    def get_signoff(self, subject=None):
        """Return a Signoff instance for this signet"""
        return self.signoff_type(signet=self, subject=subject)

    @property
    def signoff(self):
        """The Signoff instance for this signet"""
        return self.get_signoff()

    @property
    def signatory(self):
        """Return the user who signed, or AnonymousUser if signed but no signatory, None if not yet signed"""
        return self.user if self.user else AnonymousUser() if self.is_signed() else None

    def sign(self, user):
        """Sign unsigned signet for given user. If self.is_signed() raises PermissionDenied"""
        if not self.is_signed():
            self.user = user
        else:
            raise PermissionDenied(f"Attempt to sign signed Signet {self}")

    def has_user(self):
        """Return True iff this signet has a user-relation"""
        return self.user_id is not None

    def is_signed(self):
        """Return True if this Signet has a persistent representation"""
        return self.id is not None

    def is_revoked(self):
        """Return True if this Signet has been revoked"""
        return hasattr(self, "revoked")

    def has_valid_signoff(self):
        """Return True iff this Signet has a valid signoff_id"""
        from signoffs import registry

        return self.signoff_id is not None and self.signoff_id in registry.signoffs

    def can_save(self):
        """Return True iff this signet is data-complete and ready to be saved, but has not been saved before"""
        return not self.is_signed() and self.has_user() and self.has_valid_signoff()

    def update(self, defaults=False, **attrs):
        """Update instance model fields with any attrs that match by name, optionally setting only unset fields"""
        for fld in filter(lambda fld: fld not in self.read_only_fields, attrs):
            if not defaults or not getattr(self, fld, None):
                setattr(self, fld, attrs[fld])
        return self

    def get_signet_defaults(self):
        """Return dict of default field values for this signet - signet MUST have user relation!"""
        defaults = settings.SIGNOFFS_SIGNET_DEFAULTS
        return (
            get_signet_defaults(self)
            if defaults is None
            else defaults(self)
            if callable(defaults)
            else defaults
        )  # otherwise, defaults must be a dict-like object

    def set_signet_defaults(self):
        """Set default field values for this signet - signet MUST have user relation!"""
        return self.update(defaults=True, **self.get_signet_defaults())

    def validate_save(self):
        """Raise ValidationError if this Signet cannot be saved, otherwise just pass."""
        self.full_clean()
        if self.is_signed():
            raise PermissionDenied(f"Unable to re-save previously signed Signet {self}")
        elif not self.can_save():
            raise PermissionDenied(f"Unable to save Signet {self}")

    def save(self, *args, **kwargs):
        """Add a 'sigil' label if there is not one & check user has permission to save this signet"""
        self.set_signet_defaults()
        self.validate_save()
        return super().save(*args, **kwargs)

    @classmethod
    def has_object_relation(cls):
        """Return True iff this Signet class has a FK relation to some object other than user"""
        relations = [
            f
            for f in cls._meta.fields
            if isinstance(f, models.ForeignKey)
            and not issubclass(f.related_model, AbstractSignet)
            and f.name != "user"
        ]
        return bool(relations)


class AbstractRevokedSignet(models.Model):
    """
    Abstract base class for all Signet Revocation models
    A Revoked Signet is a record of a signet that was removed.  Only needed if a record of revoked signets is required.
    Persistence layer for revoked signet: who, when, what (what is an app-relative concrete Signet model)
    Note: user relation is required for signing - enforced by app logic rather than at DB level - see AbstractSignet
    """

    # noinspection PyUnresolvedReferences
    signet = models.OneToOneField(
        "Signet", on_delete=models.CASCADE, related_name="revoked"
    )
    user = models.ForeignKey(
        get_user_model(),
        on_delete=models.SET_NULL,
        null=True,
        verbose_name="Revoked by",
        related_name="+",
    )
    reason = models.TextField(blank=True, null=True)
    timestamp = models.DateTimeField(
        default=timezone.now, editable=False, null=False, verbose_name="Revoked at"
    )

    class Meta:
        abstract = True

    def __str__(self):
        return f"Revoked {getattr(self.signet, 'signoff_id', '')} by {self.user} at {self.timestamp}"



"""
    Test fixture factories for signoff models
"""
import uuid
from functools import partial

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

User = get_user_model()
auth_content_type = partial(ContentType.objects.get_for_model, User)


def get_perm(codename, name=None, content_type=None):
    """Get or create and return a permission with given codename"""
    name = name or codename.title()
    content_type = content_type or auth_content_type()
    perm, _ = Permission.objects.get_or_create(
        codename=codename,
        name=name,
        content_type=content_type,
    )
    return perm


def revoke_all_permissions(user):
    user.user_permissions.clear()


def revoke_permissions(user, *perms, content_type=None):
    content_type = content_type or auth_content_type()
    for perm in perms:
        if type(perm) is str:
            perm = get_perm(perm, content_type=content_type)
        user.user_permissions.remove(perm)


def grant_permissions(user, *perms, content_type=None):
    content_type = content_type or auth_content_type()
    for perm in perms:
        if type(perm) is str:
            perm = get_perm(perm, content_type=content_type)
        user.user_permissions.add(perm)


def get_user(
    first_name="Big",
    last_name="Bird",
    email="bigbird@example.com",
    username=None,
    password="password",
    perms=(),
    **kwargs,
):
    """
    Return a user object with given attributes and set of permissions
    """
    username = (
        username or str(uuid.uuid1())[:-10]
    )  # unique username if caller doesn't care
    user, created = User.objects.get_or_create(
        username=username,
        first_name=first_name,
        last_name=last_name,
        email=email,
        **kwargs,
    )
    if created:
        user.set_password(password)
        user.save()
    else:
        revoke_all_permissions(user)
    grant_permissions(user, *perms)
    # Add a few convenience methods to user object
    user.grant_permission = grant_permissions.__get__(user)
    user.revoke_permission = revoke_permissions.__get__(user)
    user.revoke_all_permissions = revoke_all_permissions.__get__(user)
    return user



"""
Signing Order pattern matching language. Defines the pattern for a Signing Order using Signoff Types

Pattern Matching is backed by regex_match backend (currently not replaceable, but that'd be a nice idea :-)
"""
import collections.abc
from functools import cached_property
from itertools import chain
from types import SimpleNamespace

from signoffs import registry

from .regex_match import (
    PatternMatcher,
    all_of,
    exactly_n,
    exactly_one,
    in_series,
    n_or_more,
    one_of,
    one_or_more,
    zero_or_more,
    zero_or_one,
)

#
#  Pluggable encode/decode logic for pattern and token objects
#

# Pattern objects are classes and tokens are instances of those classes (useful for testing)
obj_repr = SimpleNamespace(
    pattern_to_str=lambda obj: obj.__name__,
    pattern_from_str=lambda name: globals()[name]
    if name in globals()
    else __builtins__[name],
    to_str=lambda obj: type(obj).__name__,
)

# Default: Pattern objects are Signoff Types and tokens are signoff instances - id's are used for pattern matching
signoff_repr = SimpleNamespace(
    pattern_to_str=lambda obj: obj.id,
    pattern_from_str=lambda name: registry.signoffs.get(name),
    to_str=lambda obj: obj.signoff_id,
)


def regex_pattern(pattern, to_str):
    """Return the equivalent regex pattern matching function for given pattern"""
    # recurse nested patterns, stopping recursion when pattern is a simple object and returning its string rep.
    pattern = [
        p.regex_pattern() if isinstance(p, SigningOrderPattern) else to_str(p)
        for p in pattern
    ]
    return pattern


# Singing Order Pattern Specifiers


class SigningOrderPattern(collections.abc.Sequence):
    """
    A abstract pattern is a hierarchically nested sequence of terms used to match a concrete linear sequence of tokens
    Terms for a Signing Order are Signoff Type classes and tokens are signoff instances.

    This class adapts the Signoff pattern and match API to the implementation provided by regex_match.PatternMatcher
    """

    regex_pattern_constructor = in_series

    def __init__(self, *pattern, token_repr=signoff_repr, **kwargs):
        """Initialize with sequence of SigningOrderPattern objects (or any of its subclasses)"""
        self.pattern = pattern
        self.token_repr = token_repr
        self.kwargs = kwargs  # allow subclasses to pass arguments through to regex pattern constructors

    @cached_property
    def pattern_matcher(self):
        return PatternMatcher(self.regex_pattern())

    def regex_pattern(self):
        """Return the equivalent regex pattern matching function for this pattern"""
        # TODO: replace with type(self).regex...
        construct = self.regex_pattern_constructor.__func__  # don't bind  self.
        return construct(
            *regex_pattern(self.pattern, self.token_repr.pattern_to_str), **self.kwargs
        )

    def match(self, *tokens):
        """Returns a MatchResult object that compares iterable of tokens to this pattern"""
        token_str = " ".join(self.token_repr.to_str(s) for s in tokens)
        match = self.pattern_matcher.match(token_str)
        if match.is_valid:
            match.next = [self.token_repr.pattern_from_str(id) for id in match.next]
        return match

    def __str__(self):
        return str(self.pattern)

    def __getitem__(self, index):
        return self.pattern[index]

    def __len__(self):
        return len(self.pattern)

    def terms(self):
        """Return a flat set of pattern terms used in this pattern"""
        return set(chain(t for term in self.pattern for t in term.terms()))


class TokenPattern(SigningOrderPattern):
    """Abstract base for simple, un-nested patterns specified by a single token."""

    def terms(self):
        """Return a flat set of pattern terms used in this pattern"""
        return set(self.pattern)


class ExactlyOne(TokenPattern):
    """A pattern that is complete when there is exactly one matching token"""

    regex_pattern_constructor = exactly_one


class Optional(TokenPattern):
    """A pattern that matches zero or one optional token"""

    regex_pattern_constructor = zero_or_one


class ZeroOrMore(TokenPattern):
    """A pattern that matches zero or more matching tokens"""

    regex_pattern_constructor = zero_or_more


class OneOrMore(TokenPattern):
    """A pattern that is complete when there are one or more matching tokens"""

    regex_pattern_constructor = one_or_more


class NTokenPattern(TokenPattern):
    """Abstract base for patterns where some number, n, tokens are required"""

    def __init__(self, *pattern, n=1, **kwargs):
        super().__init__(*pattern, n=n, **kwargs)


class ExactlyN(NTokenPattern):
    """A pattern that is complete with exactly n tokens"""

    regex_pattern_constructor = exactly_n


class AtLeastN(NTokenPattern):
    """
    A pattern that is complete when pattern has at least some min. number of tokens.
    Note: match algorithm consumes all consecutive matching tokens (greedy)
    """

    regex_pattern_constructor = n_or_more


# Pattern Sets


class PatternSet(SigningOrderPattern):
    """Abstract base for nested patterns where terms are themselves patterns"""

    def __init__(self, *pattern, **kwargs):
        """Wrap any "naked" tokens in an ExactlyOne"""
        pattern = tuple(
            token if isinstance(token, SigningOrderPattern) else ExactlyOne(token)
            for token in pattern
        )
        super().__init__(*pattern, **kwargs)


class AnyOneOf(PatternSet):
    """A pattern that matches any one of a set of alternate patterns"""

    regex_pattern_constructor = one_of


class InSeries(PatternSet):
    """A pattern where tokens must be in sequential order"""

    regex_pattern_constructor = in_series


class InParallel(PatternSet):
    """
    A pattern where tokens can be in any order
    Notes:
        Experimental - only works for straight-forward, unambiguous cases.
        Terms from InSeries and AtLeastN patterns match sequential tokens, even when wrapped within InParallel
          It may be possible to handle AtLeastN terms as non-sequential, but a look-ahead match algorithm is needed?
    """

    regex_pattern_constructor = all_of


__all__ = [
    "AnyOneOf",
    "AtLeastN",
    "ExactlyN",
    "ExactlyOne",
    "InParallel",
    "InSeries",
    "OneOrMore",
    "Optional",
]



"""
Simple Regular Expression Pattern Matching language for a sequence of tokens

Design Constraints:
    API needs to be simple and clear - no regular expressions leaking through.

    Partial matches are essential & need to obtain next unmatched tokens in sequence.
"""
import itertools
from dataclasses import dataclass, field

import regex


@dataclass
class Pattern:
    """
    A pattern consists of the regular expression for matching the pattern
        and the list of tokens matched in the regular expression
    """

    regex: str
    tokens: list


#
# PATTERN CONSTRUCTORS
#


def wrap(s):
    """
    Create a Pattern out of a single string token
        this is a helper function that would usually be called by one of the other pattern constructors
        if s is a Pattern that already has a regex and token(s) associated with it, simply pass it through
        otherwise, create a regular expression chunk that matches s and record s as the associated token
        the regular expression chunk puts s into a named group, so matched instances can later be retrieved by name
    """
    if isinstance(s, str):
        escape_set = ["-", ".", "|", "(", ")", "{", "}", "+", "*", "?"]
        group_name = s
        s_pat = s
        for e in escape_set:
            s_pat = s_pat.replace(e, "\\" + e)
            group_name = group_name.replace(e, "")
        return Pattern(r"((?P<" + group_name + r">" + s_pat + r") )", [s])
    else:
        return s


def exactly_one(s):
    """Given a string token or pattern s, create a pattern that matches exactly one occurrence of s"""
    return wrap(s)


def zero_or_one(s):
    ws = wrap(s)
    return Pattern("(" + ws.regex + "{0,1})", ws.tokens)


def zero_or_more(s):
    """Optionally match zero or one of token, s"""
    ws = wrap(s)
    return Pattern("(" + ws.regex + "*)", ws.tokens)


def exactly_n(s, n):
    """Given a string token or pattern s, create a pattern that matches exactly n occurrences of s"""
    ws = wrap(s)
    return Pattern(ws.regex + "{" + str(n) + "," + str(n) + "}", ws.tokens)


def one_or_more(s):
    """Given a string token or pattern s, create a pattern that matches one or more occurrences of s"""
    ws = wrap(s)
    return Pattern("(" + ws.regex + "+)", ws.tokens)


def n_or_more(s, n):
    """Given a string token or pattern s, create a pattern that matches n or more occurrences of s"""
    ws = wrap(s)
    return Pattern("(" + ws.regex + "{" + str(n) + ",})", ws.tokens)


def in_series(*lst):
    """Given a list lst with string tokens or patterns, create a pattern that matches the given list in order"""
    p = []
    tokens = []
    for e in lst:
        ws = wrap(e)
        p = p + [ws.regex]
        tokens = tokens + ws.tokens
    return Pattern("(" + "".join(p) + ")", tokens)


def all_of(*lst):
    """
    Given a list with string tokens or patterns, create a pattern that matches if and only if
    all of the elements of lst occur, in no particular order.
    Since there is no AND operator in regular expressions, the only way to implement this that doesn't
    potentially have unintended side effects on how the rest of the expression is matched
    is to chain all permutations with an OR - ugh!
    """
    p = []
    tokens = []
    for e in lst:
        ws = wrap(e)
        p = p + [ws.regex]
        tokens = tokens + ws.tokens
    permutations = list(itertools.permutations(p))
    return Pattern("(" + "|".join(list(map("".join, permutations))) + ")", tokens)


def one_of(*lst):
    """Given a list with string tokens or patterns, create a pattern that matches any one of the patterns/tokens"""
    p = []
    tokens = []
    for e in lst:
        ws = wrap(e)
        p = p + [ws.regex]
        tokens = tokens + ws.tokens
    return Pattern("(" + "|".join(p) + ")", tokens)


@dataclass
class MatchResult:
    """Defines the result from matching a pattern template against a concrete input sequence"""

    is_valid: bool = False  # True iff string does not violate pattern, but may only be a partial match
    is_complete: bool = False  # True iff the string is a complete match for pattern
    matched: dict = field(default_factory=lambda: {})  # a dict of matched tokens
    next: list = field(
        default_factory=lambda: []
    )  # list of token(s) that would match next


class PatternMatcher:
    """
    A special-purpose pattern matcher for a sequence of tokens
    Must be initialised with a Pattern template (see Pattern class above) to be matched against
        - the Pattern template would usually be produced by utilizing a combination
          of the pattern generator functions defined above
    """

    def __init__(self, pattern):
        self.template = regex.compile(pattern.regex)
        # extract and store unique tokens in the order they appear in the pattern
        self.tokens = list(dict.fromkeys(pattern.tokens))

    def match(self, token_str):
        """Match the string of concrete tokens against this matcher's pattern, return a MatchResult"""
        # add a space at the end if there isn't one - makes life easier if we can rely on that space always being there
        token_str = token_str + (
            "" if token_str.endswith(" ") or not token_str else " "
        )
        at_least_one_token = len(token_str) > 1
        regex_match = self.template.fullmatch(token_str, partial=True)
        if regex_match or not at_least_one_token:
            next = []
            for t in self.tokens:  # find tokens that may appear next in pattern
                if self.template.fullmatch(token_str + t + " ", partial=True):
                    next.append(t)
            return MatchResult(
                is_valid=True,
                is_complete=at_least_one_token and not regex_match.partial,
                matched=regex_match.capturesdict() if at_least_one_token else {},
                next=next,
            )
        else:
            return MatchResult()



"""
App-independent tests for Approval models - no app logic
"""
from django.core import exceptions
from django.test import SimpleTestCase, TestCase

import signoffs.core.signing_order as so
from signoffs.core.approvals import ApprovalLogic, BaseApproval
from signoffs.registry import approvals, register

from . import fixtures
from .models import ApprovalSignoff, LeaveApproval, OtherStamp, Stamp


@register(id="signoffs.tests.my_approval")
class MyApproval(BaseApproval):
    stampModel = Stamp
    label = "Test Approval"

    first_signoff = ApprovalSignoff.register(id="test.approval.first")
    second_signoff = ApprovalSignoff.register(id="test.approval.second")
    final_signoff = ApprovalSignoff.register(id="test.approval.final")

    signing_order = so.SigningOrder(
        first_signoff, so.AtLeastN(second_signoff, n=2), final_signoff
    )


class SimpleApprovalTypeTests(SimpleTestCase):
    def test_approval_type_relations(self):
        approval_type = approvals.get("signoffs.tests.my_approval")
        approval = approval_type()
        self.assertEqual(approval.stamp_model, Stamp)
        stamp = approval.stamp
        self.assertEqual(stamp.approval_type, approval_type)
        self.assertEqual(stamp.approval, approval)

    def test_with_no_stamp(self):
        with self.assertRaises(exceptions.ImproperlyConfigured):
            MyApproval.register(id="test.invalid.no_stamp", stampModel=None)


class ApprovalTypeIntheritanceTests(SimpleTestCase):
    def test_class_var_overrides(self):
        a = MyApproval.register("signoff.test.my_approval.test1")
        self.assertEqual(a.label, MyApproval.label)
        self.assertEqual(a().stamp_model, Stamp)

    def test_field_override(self):
        a = MyApproval.register(
            "signoff.test.my_approval.test2",
            label="Something",
            stampModel=OtherStamp,
            logic=ApprovalLogic(revoke_perm="auth.some_perm"),
        )
        self.assertEqual(a.label, "Something")
        self.assertEqual(a.logic.revoke_perm, "auth.some_perm")
        self.assertEqual(a().stamp_model, OtherStamp)


class ApprovalTypeTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.restricted_user = fixtures.get_user(username="restricted")
        cls.approving_user = fixtures.get_user(
            username="approving", perms=("some_perm",)
        )
        cls.unrestricted_user = fixtures.get_user(
            username="permitted", perms=("some_perm", "revoke_perm")
        )

    def test_init(self):
        stamp = Stamp(approval_id=MyApproval.id)
        a1 = MyApproval(stamp=stamp)  # with explicit stamp
        self.assertEqual(a1.stamp, stamp)
        self.assertEqual(stamp.get_approval(), a1)

        a2 = MyApproval()  # default approval
        self.assertTrue(isinstance(a2.stamp, Stamp))
        self.assertEqual(a2.stamp.approval, a2)
        self.assertEqual(a2.stamp.approval_id, MyApproval.id)
        self.assertFalse(a2.stamp.is_approved())

    def test_invalid_init(self):
        a1 = MyApproval.register(
            "signoff.test.my_approval.test3",
            stampModel=OtherStamp,
            logic=ApprovalLogic(revoke_perm="some_perm"),
        )
        stamp = Stamp(approval_id=MyApproval.id)
        with self.assertRaises(exceptions.ImproperlyConfigured):
            a1(stamp=stamp)  # stamp model does not match approval
        a = a1()
        self.assertFalse(
            a.can_revoke(user=self.restricted_user)
        )  # approval requires permission
        with self.assertRaises(exceptions.PermissionDenied):
            a.revoke_if_permitted(user=self.restricted_user)

    def test_create(self):
        a = MyApproval.create()
        self.assertFalse(a.is_approved())


class SigningOrderTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.restricted_user = fixtures.get_user(username="restricted")
        cls.approving_user = fixtures.get_user(
            username="approving", perms=("some_perm",)
        )
        cls.unrestricted_user = fixtures.get_user(
            username="permitted", perms=("some_perm", "revoke_perm")
        )
        cls.approval = MyApproval().save()

    def test_next_signoffs(self):
        u = self.unrestricted_user
        next = self.approval.next_signoffs(for_user=u)
        self.assertEqual(len(next), 1)
        self.assertEqual(next[0].id, MyApproval.first_signoff.id)
        self.assertEqual(next[0].signet.stamp, self.approval.stamp)
        self.assertTrue(next[0].can_sign(user=u))
        next[0].sign(user=u)
        self.assertEqual(self.approval.signatories.count(), 1)
        next = self.approval.next_signoffs(for_user=u)
        self.assertEqual(len(next), 1)
        self.assertEqual(next[0].id, MyApproval.second_signoff.id)

    def test_can_sign(self):
        u = self.unrestricted_user
        self.assertTrue(self.approval.can_sign(user=u))

    def test_is_complete(self):
        self.assertFalse(self.approval.is_complete())
        u = self.unrestricted_user
        self.approval.next_signoffs(for_user=u)[0].sign(user=u)
        self.approval.next_signoffs(for_user=u)[0].sign(user=u)
        u2 = self.approving_user
        self.approval.next_signoffs(for_user=u2)[0].sign(user=u)
        self.assertFalse(self.approval.is_complete())

        next = self.approval.next_signoffs(for_user=u)
        self.assertEqual(len(next), 2)
        final = [s for s in next if s.id == MyApproval.final_signoff.id][0]
        final.sign(user=u)
        self.assertTrue(self.approval.is_complete())


class ApprovalTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = fixtures.get_user()
        cls.approval = MyApproval().save()

    def test_next_signoffs(self):
        u = self.user
        # One first_signoff
        next = self.approval.next_signoffs(for_user=u)
        self.assertEqual(next[0].id, MyApproval.first_signoff.id)
        next[0].sign(user=u)
        # Two second_signoffs
        next = self.approval.next_signoffs(for_user=u)
        self.assertEqual(next[0].id, MyApproval.second_signoff.id)
        next[0].sign(user=u)
        next = self.approval.next_signoffs(for_user=u)
        self.assertEqual(next[0].id, MyApproval.second_signoff.id)
        next[0].sign(user=u)
        # final signoff now available
        next = self.approval.next_signoffs(for_user=u)
        self.assertSetEqual(
            {s.id for s in next},
            {MyApproval.second_signoff.id, MyApproval.final_signoff.id},
        )
        self.assertFalse(self.approval.is_complete() or self.approval.is_approved())
        next[-1].sign(user=u)
        self.assertTrue(self.approval.is_complete())
        self.assertFalse(self.approval.is_approved())
        self.approval.approve_if_ready()
        self.assertTrue(self.approval.is_approved())

    def sign_all(self, user):
        """Complete signatures on self.approval with given user, return list of signoffs made"""
        s = []
        next = True
        while not self.approval.is_complete() and next:
            next = self.approval.next_signoffs(for_user=user)
            s += [so.sign(user=user) for so in next]
        return s

    def test_revoke(self):
        u = self.user
        signs = self.sign_all(u)
        # Only last signoff made can be revoked
        self.assertTrue(self.approval.can_revoke_signoff(signs[-1], u))
        self.assertFalse(self.approval.can_revoke_signoff(signs[-2], u))
        self.assertFalse(self.approval.can_revoke(u))

        self.approval.approve_if_ready()
        self.assertTrue(self.approval.is_approved())

        self.assertFalse(self.approval.can_revoke_signoff(signs[-1], u))
        self.assertTrue(self.approval.can_revoke(u))

        self.approval.revoke(u)
        # all signoffs are also revoked
        self.assertFalse(self.approval.is_approved())
        self.assertEqual(self.approval.signoffs.count(), 0)

    def test_contains(self):
        u = self.user
        self.sign_all(u)
        self.assertTrue(u in self.approval)
        self.assertFalse(fixtures.get_user() in self.approval)
        self.assertTrue(MyApproval.first_signoff in self.approval)
        self.assertTrue(MyApproval.final_signoff in self.approval)
        self.assertTrue(MyApproval.first_signoff() in self.approval)
        self.assertTrue(self.approval.signoffs.all()[0] in self.approval)
        self.assertFalse(ApprovalSignoff() in self.approval)
        self.assertTrue(MyApproval.first_signoff.id in self.approval)
        self.assertFalse("no.such.signoff" in self.approval)


class ApprovalSignoffTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.unrestricted_user = fixtures.get_user(
            username="permitted", perms=("some_perm", "revoke_perm")
        )
        cls.approval = MyApproval().save()

    def sign_all(self):
        u = self.unrestricted_user
        self.approval.next_signoffs(for_user=u)[0].sign(user=u)
        self.approval.next_signoffs(for_user=u)[0].sign(user=u)
        self.approval.next_signoffs(for_user=u)[0].sign(user=u)
        self.approval.next_signoffs(for_user=u)[-1].sign(user=u)

    def test_subject_relation_populated(self):
        # start with a approval that has a subject relation
        self.approval._subject = object()
        self.assertIsNotNone(self.approval.subject)
        # signoffs on this approval will have the approval as their subject...
        signoff = self.approval.get_next_signoff()
        self.assertEqual(signoff.signet.stamp, self.approval.stamp)
        self.assertEqual(signoff.subject, self.approval)
        # and the relation persists so long as we have the exact same subject...
        self.assertEqual(signoff.subject.subject, self.approval.subject)
        # But, signet stamp's approval is equal, but not the same object!!  Access via stamp constructs new instance!
        self.assertEqual(self.approval, signoff.signet.stamp.approval)
        self.assertIsNot(self.approval, signoff.signet.stamp.approval)
        # With the implication... (not a desired design goal, but a consequence of the architecture)
        signets_signoff = (
            signoff.signet.signoff
        )  # acquire signoff from signet, e.g., as you might in a view, leads to:
        self.assertIsNone(signets_signoff.subject)  # in other words...
        self.assertNotEqual(signets_signoff.subject, self.approval)

    def test_subject_relation_on_signoffs_manager(self):
        # start with a bunch of signoffs signed
        self.sign_all()
        signoffset = self.approval.signoffs.all()
        self.assertTrue(all(s.subject == self.approval for s in signoffset))
        earliest = self.approval.signoffs.earliest()
        self.assertTrue(earliest.subject == self.approval)
        latest = self.approval.signoffs.latest()
        self.assertTrue(latest.subject == self.approval)


class ApprovalQuerysetTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.myapprovals = [
            MyApproval.create(),
            MyApproval.create(approved=True),
            MyApproval.create(),
        ]
        cls.leaveapprovals = [
            LeaveApproval.create(approved=True),
            LeaveApproval.create(),
        ]

    def test_stamp_queryset(self):
        myapproval_qs = MyApproval.get_stamp_queryset().approvals()
        self.assertListEqual(myapproval_qs, self.myapprovals)
        leaveapproval_qs = LeaveApproval.get_stamp_queryset().approvals()
        self.assertListEqual(leaveapproval_qs, self.leaveapprovals)

    def test_stamp_queryset_filter(self):
        approved_qs = MyApproval.get_stamp_queryset().filter(approved=True).approvals()
        self.assertListEqual(
            approved_qs, [a for a in self.myapprovals if a.is_approved()]
        )


class StampModelTests(TestCase):
    def test_valid_approval_type(self):
        a = approvals.get("signoffs.tests.my_approval")
        p = Stamp(approval_id="signoffs.tests.my_approval")
        self.assertEqual(p.approval_type, a)

    def test_invalid_approval_type(self):
        p = Stamp(approval_id="not.a.valid.type")
        with self.assertRaises(exceptions.ImproperlyConfigured):
            self.assertFalse(p.approval_type)

    def test_signatories(self):
        p = Stamp(approval_id="signoffs.tests.my_approval")
        p.save()
        u = fixtures.get_user(username="daffyduck")
        p.signatories.create(user=u, stamp=p, signoff_id="test.approval.first")
        self.assertTrue(p.is_user_signatory(u))

    def approve(self):
        p = Stamp(approval_id="signoffs.tests.my_approval")
        p.approve()
        p.save()
        self.assertTrue(p.is_approved())


class StampQuerysetTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        u = fixtures.get_user()
        set1 = (
            MyApproval.create(),
            MyApproval.create(),
        )
        set2 = (
            LeaveApproval.create(),
            LeaveApproval.create(),
            LeaveApproval.create(),
        )
        cls.all_approvals = set1 + set2
        cls.approval_set1, cls.approval_set2 = set1, set2
        cls.user = u

    def test_qs_basics(self):
        approvals = Stamp.objects.filter(approval_id=MyApproval.id)
        self.assertQuerysetEqual(
            approvals.order_by("pk"), [a.stamp for a in self.approval_set1]
        )

    def test_qs_approvals(self):
        approvals = MyApproval.get_stamp_queryset().approvals()
        self.assertQuerysetEqual(approvals, self.approval_set1)

    def test_qs_approvals_filter(self):
        base_qs = Stamp.objects.order_by("pk")
        self.assertQuerysetEqual(
            base_qs.approvals(approval_id=MyApproval.id), self.approval_set1
        )
        self.assertQuerysetEqual(
            base_qs.approvals(approval_id=LeaveApproval.id), self.approval_set2
        )

    def test_qs_approvals_performance(self):
        base_qs = Stamp.objects.all().order_by("pk")
        with self.assertNumQueries(1):
            approvals1 = base_qs.approvals(approval_id=MyApproval.id)
            self.assertEqual(len(approvals1), len(self.approval_set1))
            approvals2 = base_qs.approvals(approval_id=LeaveApproval.id)
            self.assertEqual(len(approvals2), len(self.approval_set2))
            self.assertEqual(len(base_qs.approvals()), len(self.all_approvals))



"""
App-independent tests for Signoff models - no app logic
"""
from django.contrib.auth import get_user_model
from django.core import exceptions
from django.test import SimpleTestCase, TestCase

from signoffs import registry
from signoffs.signoffs import SignoffLogic

from . import fixtures
from .models import BasicSignoff, OtherSignet, Signet

signoff1 = BasicSignoff.register(id="test.signoff1")
signoff2 = BasicSignoff.register(
    id="test.signoff2",
    signetModel=OtherSignet,
    label="Something",
    logic=SignoffLogic(perm="auth.some_perm", revoke_perm="auth.revoke_perm"),
)
signoff3 = BasicSignoff.register(id="test.signoff3")


class SimpleSignoffTypeTests(SimpleTestCase):
    def test_signoff_type_relations(self):
        signoff_type = registry.signoffs.get("test.signoff1")
        signoff = signoff_type()
        self.assertEqual(signoff.signet_model, Signet)
        signet = signoff.signet
        self.assertEqual(signet.signoff_type, signoff_type)
        self.assertEqual(signet.signoff, signoff)

    def test_with_no_signet(self):
        with self.assertRaises(exceptions.ImproperlyConfigured):
            BasicSignoff.register(id="test.invalid.no_signet", signetModel=None)


class SignoffTypeIntheritanceTests(SimpleTestCase):
    def test_class_var_overrides(self):
        s = registry.signoffs.get("test.signoff1")
        self.assertEqual(s.label, BasicSignoff.label)
        self.assertEqual(s().signet_model, Signet)

    def test_field_override(self):
        so = registry.signoffs.get("test.signoff2")
        s = so()
        self.assertEqual(s.label, "Something")
        self.assertEqual(s.logic.perm, "auth.some_perm")
        self.assertEqual(s.signet_model, OtherSignet)


class SignoffTypeTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.restricted_user = fixtures.get_user(username="restricted")
        cls.signing_user = fixtures.get_user(username="signing", perms=("some_perm",))
        cls.unrestricted_user = fixtures.get_user(
            username="permitted", perms=("some_perm", "revoke_perm")
        )

    def test_init(self):
        signet = Signet(signoff_id=signoff1.id)
        so1 = signoff1(signet=signet)  # with explicit signet
        self.assertEqual(so1.signet, signet)

        so2 = signoff1()  # default, unsigned signoff
        self.assertTrue(isinstance(so2.signet, Signet))
        self.assertEqual(so2.signet.signoff_id, signoff1.id)
        self.assertFalse(so2.signet.has_user())

        so3 = signoff1(
            user=self.signing_user, sigil="custom sigil"
        )  # with signet construction args
        self.assertEqual(so3.signet.user, self.signing_user)
        self.assertEqual(so3.signet.sigil, "custom sigil")

    def test_invalid_init(self):
        signet = Signet(signoff_id=signoff1.id)
        with self.assertRaises(exceptions.ImproperlyConfigured):
            signoff1(
                signet=signet, user=self.signing_user
            )  # supply either signet OR signet create kwargs
        with self.assertRaises(exceptions.ImproperlyConfigured):
            signoff2(
                signet=signet, user=self.signing_user
            )  # signet model does not match signoff
        so = signoff2(user=self.restricted_user)  # signoff requires permission
        self.assertFalse(so.can_save())
        with self.assertRaises(exceptions.PermissionDenied):
            so.save()

    def test_can_sign(self):
        unrestricted_so = signoff1()
        restricted_so = signoff2()
        self.assertTrue(unrestricted_so.can_sign(self.restricted_user))
        self.assertFalse(restricted_so.can_sign(self.restricted_user))
        self.assertTrue(restricted_so.can_sign(self.signing_user))

    def test_can_revoke(self):
        unrestricted_so = signoff1(user=self.unrestricted_user).save()
        restricted_so = signoff2(user=self.unrestricted_user).save()
        self.assertTrue(unrestricted_so.can_revoke(self.restricted_user))
        self.assertFalse(restricted_so.can_revoke(self.restricted_user))
        self.assertFalse(restricted_so.can_revoke(self.signing_user))
        self.assertTrue(restricted_so.can_revoke(self.unrestricted_user))

    def test_irrevokable(self):
        signoff = BasicSignoff.register(
            id="test.irrevokable", logic=SignoffLogic(revoke_perm=False)
        )
        self.assertFalse(signoff.is_permitted_revoker(self.unrestricted_user))
        irrevokable_so = signoff(user=self.signing_user).save()
        self.assertFalse(irrevokable_so.can_revoke(self.unrestricted_user))

    def test_create(self):
        so = signoff2.create(user=self.signing_user)
        self.assertTrue(so.is_signed)

    def test_default_sigil(self):
        so = signoff2.create(user=self.signing_user)
        self.assertEqual(so.signet.sigil, self.signing_user.get_full_name())


class SignoffQuerysetTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        u = fixtures.get_user(perms=("some_perm",))
        cls.user = u
        u2 = fixtures.get_user(perms=("some_perm",))

        cls.signoff1s = [
            signoff1.create(
                user=u,
            ),
            signoff1.create(
                user=u2,
            ),
            signoff1.create(
                user=u,
            ),
        ]
        cls.signoff2s = [
            signoff2.create(
                user=u2,
            ),
            signoff2.create(
                user=u,
            ),
        ]
        cls.signoff3s = [
            signoff3.create(
                user=u,
            ),
        ]

    def test_signet_queryset(self):
        so1_qs = signoff1.get_signet_queryset().signoffs()
        self.assertListEqual(so1_qs, self.signoff1s)
        so2_qs = signoff2.get_signet_queryset().signoffs()
        self.assertListEqual(so2_qs, self.signoff2s)
        so3_qs = signoff3.get_signet_queryset().signoffs()
        self.assertListEqual(so3_qs, self.signoff3s)

    def test_signet_queryset_filter(self):
        so_qs = signoff1.get_signet_queryset().filter(user=self.user).signoffs()
        self.assertListEqual(
            so_qs, [so for so in self.signoff1s if so.signet.user == self.user]
        )

    def test_signoff_get_basic(self):
        self.assertEqual(signoff2.get(user=self.user), self.signoff2s[1])
        self.assertIsNone(signoff2.get(user=fixtures.get_user()).signet.pk)
        with self.assertRaises(exceptions.MultipleObjectsReturned):
            signoff1.get(user=self.user)

    def test_signoff_get_by_type(self):
        """test that signoff.get only retrieves signoffs of right type even when they share same Signet model"""
        signoffa = signoff1.register(id="test.signoffa")
        signoffb = signoff1.register(id="test.signoffb")
        other_user = fixtures.get_user()
        a1 = signoffa.create(user=self.user)
        a2 = signoffa.create(user=other_user)
        b1 = signoffb.create(user=self.user)
        b2 = signoffb.create(user=other_user)
        self.assertEqual(signoffa.get(user=self.user), a1)
        self.assertEqual(signoffa.get(user=other_user), a2)
        self.assertEqual(signoffb.get(user=self.user), b1)
        self.assertEqual(signoffb.get(user=other_user), b2)

    def test_signoff_get_with_queryset(self):
        qs = signoff1.get_signetModel().objects.filter(pk=1)
        self.assertEqual(signoff1.get(queryset=qs, user=self.user), self.signoff1s[0])


class SignetModelTests(SimpleTestCase):
    def test_default_signature(self):
        u = get_user_model()(username="daffyduck")
        o = Signet(signoff_id="test.signoff1", user=u)
        self.assertEqual(o.get_signet_defaults()["sigil"], "daffyduck")
        u = get_user_model()(username="daffyduck", first_name="Daffy", last_name="Duck")
        o = Signet(signoff_id="test.signoff1", user=u)
        self.assertEqual(o.get_signet_defaults()["sigil"], "Daffy Duck")

    def test_valid_signoff_type(self):
        s = registry.signoffs.get("test.signoff1")
        o = Signet(signoff_id="test.signoff1")
        self.assertEqual(o.signoff_type, s)

    def test_invalid_signoff_type(self):
        o = Signet(signoff_id="not.a.valid.type")
        with self.assertRaises(exceptions.ImproperlyConfigured):
            self.assertFalse(o.signoff_type)


class SignetQuerysetTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        user = fixtures.get_user(
            first_name="Daffy", last_name="Duck", username="quacker"
        )
        signoff1_set = (
            signoff1.create(user=user),
            signoff1.create(user=user),
        )
        signoff3_set = (
            signoff3.create(user=user),
            signoff3.create(user=user),
            signoff3.create(user=user),
        )
        cls.signoff1_set = signoff1_set
        cls.signoff3_set = signoff3_set
        cls.all_signoffs = signoff1_set + signoff3_set
        cls.user = user

    def test_qs_basics(self):
        signoff_set = Signet.objects.filter(signoff_id="test.signoff1")
        self.assertQuerysetEqual(
            signoff_set.order_by("pk"), [so.signet for so in self.signoff1_set]
        )

    def test_qs_signoffs(self):
        signoff_set = (
            Signet.objects.order_by("pk").filter(signoff_id="test.signoff1").signoffs()
        )
        self.assertQuerysetEqual(signoff_set, self.signoff1_set)

    def test_qs_signoffs_filter(self):
        base_qs = Signet.objects.order_by("pk")
        self.assertQuerysetEqual(
            base_qs.signoffs(signoff_id="test.signoff1"), self.signoff1_set
        )
        self.assertQuerysetEqual(base_qs.signoffs(signoff_id="test.signoff2"), [])
        self.assertQuerysetEqual(
            base_qs.signoffs(signoff_id="test.signoff3"), self.signoff3_set
        )

    def test_qs_signoffs_performance(self):
        base_qs = Signet.objects.all().order_by("pk")
        with self.assertNumQueries(1):
            signoffs1 = base_qs.signoffs(signoff_id="test.signoff1")
            self.assertEqual(len(signoffs1), len(self.signoff1_set))
            signoffs2 = base_qs.signoffs(signoff_id="test.signoff2")
            self.assertEqual(len(signoffs2), 0)
            signoffs3 = base_qs.signoffs(signoff_id="test.signoff3")
            self.assertEqual(len(signoffs3), len(self.signoff3_set))
            self.assertEqual(len(base_qs.signoffs()), len(self.all_signoffs))



"""
    Classes, Descriptors, and decorators for coordinating state transitions driven by an Approval Process
    Used to manage a multi-state approval process where state transitions are triggered by Approvals.
    Core responsibility: ensure integrity / consistency of approval state and state transitions

    A "transition" is just a method on your process model that manages a single state transition.
    An "approval transition" is a transition that is dependent on an Approval being approved or revoked.

    Transitions handle side-effects.  Generally, decorators, like django_fsm.transition, are used to perform
        any state transitions (like approvals or revokes), and like django_fsm, these state changes must be
        saved to the DB as a separate step after a successful transition is made.
        See convenience methods BasicApprovalProcess.try_*_transition for examples of how to correctly compelte a transition.

    django-fsm integration:
        - FSMApprovalProcess enforces FSM logic withing the BasicApprovalProcess API
        - FSMApprovalProcessDescriptor provides a declarative syntax for defining an FSM Approval Process
"""
import inspect
from collections import defaultdict
from dataclasses import dataclass, field
from functools import wraps
from typing import Any, Callable, Dict

from django.core.exceptions import ImproperlyConfigured
from django.db import models, transaction

from signoffs import registry
from signoffs.core import approvals, renderers


@dataclass
class ApprovalTransition:
    """
    Associate an Approval Type with callables representing transitions to take when approval actions are performed.
    When used with  BasicApprovalProcess, transition callables must take 2 arguments:
        - process_model instance (usually named self on callable)
        - approval instance (the approval that is driving the transition
    """

    approval_id: str = None
    approve: Callable = None
    revoke: Callable = None

    @property
    def approve_name(self):
        return self.approve.__name__ if self.approve else ""

    @property
    def revoke_name(self):
        return self.revoke.__name__ if self.revoke else ""


@dataclass
class ApprovalTransitionRegistry:
    """A registry associating Approval Types with the transition function that implement approval actions"""

    transitions: Dict = field(default_factory=lambda: defaultdict(ApprovalTransition))

    def __iter__(self):
        return iter(self.transitions.values())

    def add_approval(self, approval_type):
        """Add given approval_type (or approval instance, or str id) to registry with no transitions (yet)"""
        approval_id = registry.get_approval_id(approval_type)

        if approval_id not in self.transitions:
            self.transitions[approval_id] = ApprovalTransition(approval_id=approval_id)

    def add_approve_transition(self, approval_type, transition):
        """Add a transition function that approves the given approval_type (or approval instance, or str id)"""
        approval_id = registry.get_approval_id(approval_type)

        t = self.transitions[approval_id]
        t.approval_id = approval_id  # in case this is a new entry
        t.approve = transition

    def add_revoke_transition(self, approval_type, transition):
        """Add a transition that revokes for given approval_type (or approval instance, or str id)"""
        approval_id = registry.get_approval_id(approval_type)

        t = self.transitions[approval_id]
        t.approval_id = approval_id  # in case this is a new entry
        t.revoke = transition

    def get(self, approval_type):
        """Return the Approval Transition associated with the given approval_type (or ...), or None"""
        approval_id = registry.get_approval_id(approval_type)
        # directly accesing self.transitions.approval_id would create a default entry!
        return self.transitions.get(approval_id, None)

    def approval_order(self):
        """Return a list of approval id's, in the order they were added to the registry"""
        return list(self.transitions.keys())


class BoundApprovalSequence(dict):
    """
    An (ordered) dictionary of Approval instances bound to an object, like an Approval Process, keyed by attribute name.
    Designed to work the approvals declared using an ApprovalField or RelatedApprovalDescriptor - for use in other
      cases, be sure type(obj) declares ApprovalType using same attribute name as approval instance on obj.
    Sequenced by an optional linear ordering, or left to app logic to define the ordering.
    Without an explicit ordering, ordering is based on introspection for declared Approvals,
        which come ordered alphabetically NOT in the order in which they are declared (ugh).
    """

    def __init__(self, obj, ordering):
        """
        Define the linear approval sequence as a list of Approval instances bound given object
        Ordering may by an iterable of Approval Types, Fields, or field names.
        """
        approval_items = self._get_ordered_items(obj, ordering)
        super().__init__(
            (name, getattr(obj, name)) for name, _ in approval_items
        )  # bound instance approvals!
        self.instance = obj
        self.is_ordered = bool(ordering)

    @classmethod
    def _get_approval_members(cls, obj):
        """Return a list of 2-tuples (name, approval type) for Approval Types / Fields defined on type(obj)"""

        def is_approval_field(attr):
            return type(attr) is type and issubclass(attr, approvals.AbstractApproval)

        # MUST inspect obj Type - accessing obj members evaluates descriptors, potentially causing inf. recursion.
        return inspect.getmembers(type(obj), predicate=is_approval_field)

    def _get_ordered_items(self, obj, ordering):
        """
        Return an iterable of 2-tuples, (name, approval type), one for each approval, ordered by ordering, if given.
        Raise ImproperlyConfigured if an ordering is provided but does not match the approvals bound to object
        """
        approval_members = self._get_approval_members(obj)
        if not ordering:
            return approval_members  # no ordering - approvals will be in alphabetic-order by attr name

        approval_order = [
            a.approval_type
            if hasattr(a, "approval_type")
            else registry.get_approval_type(a)
            for a in ordering
        ]

        if not set(a.id for a in approval_order).issubset(
            set(a[1].id for a in approval_members)
        ):
            raise ImproperlyConfigured(
                f"Ordering {approval_order} is inconsistent with approvals {approval_members} declared on {obj}"
            )
        name = {approval: name for name, approval in approval_members}
        return ((name[approval], approval) for approval in approval_order)


@dataclass
class AbstractPersistTransition:
    """
    Callable to save objects modified by a transition to the DB (and/or perform other follow-up actions)
    BasicApprovalProcess.try_*_transition methods call it on successful transition.
    A transition fn decorated with BasicApprovalProcess.do_approval or .do_revoke will return an appropriate PersistTransition object.
    """

    instance: models.Model
    approval: approvals.AbstractApproval
    result: Any

    def __call__(self, *args, **kwargs):
        """Do whatever is needed to persist transition"""
        raise NotImplementedError


class TransactionSave(AbstractPersistTransition):
    """Callable to save modified objects in a transaction to maintain data integrity"""

    def __call__(self, *args, **kwargs):
        """Save the transitioned instance and the approved approval in a transaction"""
        with transaction.atomic():  # Do both the approval and state transition together, or do neither.
            self.approval.save()
            self.instance.save()


@dataclass
class TransactionRevoke(AbstractPersistTransition):
    """Callable to revoke the approval and save modified objects in a transaction to maintain data integrity"""

    def __call__(self, user, *args, **kwargs):
        """Save the transitioned instance and revoke the approval for given user in a transaction"""
        with transaction.atomic():  # Do both the revoke and state transition together, or do neither.
            self.approval.revoke(user=user)
            self.instance.save()


# TODO: define ApprovalProcess Protocol


class BasicApprovalProcess:
    """
    Associate approval actions in a multi-approval process with functions (transitions) that trigger those actions
    Ensure actions preserve integrity of approval process state, keeping approval state and process state in sync.
    Most usefully defined on an Approval Process Model class using the ApprovalProcessDescriptor,
        which provides decorators to define and register the transitions.
    """

    approval_sequence_class = BoundApprovalSequence
    transition_save_class = TransactionSave
    transition_revoke_class = TransactionRevoke

    render: renderers.ApprovalProcessRenderer = (
        renderers.ApprovalProcessRenderer()
    )  # presentation logic service

    def __init__(self, process_model, transition_registry, approval_sequence=None):
        """
        Associate the transition registry with the approval process_model instance on which the transitions are defined.
        Process model is a class that typically defines a set of ApprovalField's and transition functions.
        Transition callables in the registry should be methods on the approval process_model, and
            take 2 positional arguments, e.g., make_transition(self, approval)
        """
        self.process_model = process_model
        self.registry = transition_registry
        self._validate_registry()
        self.seq = approval_sequence or self.approval_sequence_class(
            process_model, ordering=self.registry.approval_order()
        )
        assert self.seq.is_ordered  # Actions always have ordered sequence

    def _validate_registry(self):
        """Each transition must be a callable method on the approval process instance."""
        for transition in self.registry:
            for action in (
                a for a in (transition.approve, transition.revoke) if a is not None
            ):
                if not (
                    hasattr(self.process_model, action.__name__) and callable(action)
                ):
                    raise ImproperlyConfigured(
                        "Transitions must be callable methods on the Approval Process object:"
                        f" {self.process_model}.{action.__name__} does not exist or is not callable."
                    )

    # access to approval transition sequencing

    def get_all_approvals(self):
        """Return list of all Approval instances defined in this sequence for the process_model"""
        return list(self.seq.values())

    def get_approved_approvals(self):
        """Return list of all approved Approvals in this sequence."""
        return [a for a in self.seq.values() if a.is_approved()]

    def get_unapproved_approvals(self):
        """Return list of all un-approved Approvals in this."""
        return [a for a in self.seq.values() if not a.is_approved()]

    def get_next_approval(self):
        """Return the "next" approval in sequence ready for signing - sensible only for ordered approvals"""
        approvals = self.get_unapproved_approvals()
        return approvals[0] if approvals else None

    def next_approval_is_signed(self):
        """Return True iff the "next" approval in sequence has at least one signatory"""
        approval = self.get_next_approval()
        return approval.is_signed() if approval else False

    def get_previous_approval(self):
        """Return the "last" approval in sequence that was approved - sensible only for ordered approvals"""
        approvals = self.get_approved_approvals()
        return approvals[-1] if approvals else None

    def get_available_approvals(self):
        """Return list of approvals that can proceed based on available state transitions"""
        return [
            approval for approval in self.seq.values() if self.can_proceed(approval)
        ]

    def get_next_available_approval(self):
        """Return the next approval available for signing"""
        try:
            return self.get_available_approvals()[0]
        except IndexError:
            return None

    def get_revokable_approvals(self):
        """Return list of approvals that can be revoked based on available state transitions"""
        return [
            approval for approval in self.seq.values() if self.is_revokable(approval)
        ]

    def get_next_revokable_approval(self):
        """Return the next approval available for revoking"""
        try:
            return self.get_revokable_approvals()[0]
        except IndexError:
            return None

    def is_revokable_approval(self, approval):
        """Return True iff the given approval is revokable within the process"""
        return approval in self.get_revokable_approvals()

    def contains_stamp(self, stamp_id):
        """Return True iff the given stamp pk is among those that are part of this approval process"""
        return stamp_id in (a.stamp.pk for a in self.get_all_approvals())

    # access to bound transition methods

    def bound_approve_transition(self, approval_type):
        """Return the associated approve transition as a bound method of process_model, or None"""
        t = self.registry.get(approval_type)
        has_transtion = t is not None and t.approve is not None
        return getattr(self.process_model, t.approve_name) if has_transtion else None

    def bound_revoke_transition(self, approval_type):
        """Return the associated revoke transition as a bound method of process_model, or None"""
        t = self.registry.get(approval_type)
        has_transtion = t is not None and t.revoke is not None
        return getattr(self.process_model, t.revoke_name) if has_transtion else None

    # encapsulated transition logic : conditions and processes for coordinating approval and state transitions

    # Approve transition logic:

    def can_proceed(self, approval, **kwargs):
        """
        Return True iff signoff on the approval can proceed
            -- is it next in sequence, available for signing, etc.
        """
        # proceed on any unapproved approvals if this sequence is unordered, otherwise only on the next approval in seq.
        return not approval.is_approved() and approval == self.get_next_approval()

    def has_approval_transition_perm(self, approval, user, **kwargs):
        """Returns True iff model in state allows transition from given approval by given user"""
        return True  # no special permissions associated with approve actions for non-FSM transitions

    def user_can_proceed(self, approval, user, **kwargs):
        """Return True if the user can proceed with signing the given approval (or approval name)"""
        return self.can_proceed(
            approval, **kwargs
        ) and self.has_approval_transition_perm(approval, user, **kwargs)

    def can_do_approve_transition(self, approval, user, **kwargs):
        """Return True iff all conditions are met for user to proceed with approval and make transition"""
        # possible there is a transition or not - either way, we can move ahead as non-FSM transitions have no perms.
        # don't call approval.ready_to_approve to avoid potential recursion.  Duplicate code instead :-(
        return approval.ready_to_approve() and self.user_can_proceed(
            approval, user, **kwargs
        )

    # Revoke transition logic:

    def is_revokable(self, approval, **kwargs):
        """Return True if the approval can be revoked"""
        # can revoke the last approved in seq. if the active approval has no signoffs
        # HACK: don't call approval.is_revokable here to avoid potential recursion!  Duplicate code instead :-(
        # return approval.is_revokable() and (
        return approval.is_approved() and (
            approval == self.get_previous_approval()
            and not self.next_approval_is_signed()
        )

    def has_revoke_transition_perm(self, approval, user, **kwargs):
        """Returns True iff model state allows transition revoking given approval by given user"""
        return approval.is_permitted_revoker(user)

    def user_can_revoke(self, approval, user, **kwargs):
        """Return True iff user can proceed with revoke transition triggered by given approval (or approval name)"""
        return self.is_revokable(
            approval, **kwargs
        ) and self.has_revoke_transition_perm(approval, user, **kwargs)

    def can_do_revoke_transition(self, approval, user, **kwargs):
        """Return True iff all conditions are met for user to proceed with revoking approval and make transition"""
        # possible there is a transition or not - either way, we can more ahead as non-FSM transitions have no perms.
        return self.user_can_revoke(approval, user, **kwargs)

    # Approval Transition Decorators

    @classmethod
    def do_approval(cls, transition_method):
        """
        Wrap transition_method with do_approval decorator that also
            returns a transition_save_class object to handle follow-up persistence logic
        """

        @wraps(transition_method)
        def _return_save_callable(instance, approval, *args, **kwargs):
            """Complete the transition & approval, return a callable to complete the follow-up state change saves"""
            result = transition_method(instance, approval, *args, **kwargs)
            approval.approve(commit=False)
            return cls.transition_save_class(instance, approval, result)

        return _return_save_callable

    @classmethod
    def do_revoke(cls, transition_method):
        """
        Wrap transition_method so it returns a  transition_revoke_class object to handle any follow-up persistence logic
        """

        @wraps(transition_method)
        def _return_revoke_callable(instance, approval, *args, **kwargs):
            """Complete the transition and return a callable to handle the revoke and state change saves"""
            result = transition_method(instance, approval, *args, **kwargs)
            return cls.transition_revoke_class(instance, approval, result)

        return _return_revoke_callable

    # Approval Transition Actions: attempt to do the transitions

    def try_approve_transition(self, approval, user):
        """
        Trigger the associated state transition to approve the given approval instance, for the given user, if possible
        Convenience method to wrap up 3 steps commonly done together:
            1) determine if the approval is ready to be approved and any associated state transition can be made;
            2) trigger the transition that makes the approval
            3) save the approval and any process_model state changes using this process's transition save object
        Transition function should return a transition_save_class callable to save all objects modified by transition
        Return True if the approval and state transition occurred, False otherwise.
        """
        if not self.can_do_approve_transition(approval, user):
            return False
        # Trigger the transition function that approves the approval
        transition = self.bound_approve_transition(approval)
        save = transition(approval)
        if isinstance(save, self.transition_save_class):
            save()
        else:
            save = self.transition_save_class(self.process_model, approval, save)
            save()
        return True

    def try_revoke_transition(self, approval, user):
        """
        Trigger the associated state transition to revoke the given approval instance, for the given user, if possible
        Convenience method to wrap up 3 steps commonly done together:
            1) determine if the approval can be revoked by user and any associated state transition can be made;
            2) trigger the transition that goes with revoking the approval
            3) revoke the approval and save any process_model state changes, using this process's transition revoke
        Transition function should return a transition_revoke_class callable to save all objects modified by transition
        Return True if the revoke and state transition occurred, False otherwise.
        """
        if not self.can_do_revoke_transition(approval, user):
            return False
        # Trigger the transition function that revokes the approval
        transition = self.bound_revoke_transition(approval)
        revoke = transition(approval)
        if isinstance(revoke, self.transition_revoke_class):
            revoke(user)
        else:
            revoke = self.transition_revoke_class(self.process_model, approval, revoke)
            revoke(user)
        return True


class FsmApprovalProcess(BasicApprovalProcess):
    """
    An ApprovalProcss that associate approval actions (approve & revoke)
        with FSM state transitions (@transition decorated functions).
    Most usefully defined on an Approval Process Model class using the FsmApprovalProcessDescriptor,
        which provides decorators to define and register the approval actions on FSM transitions.
    Assumes that all transitions are FSM transitions and all are registered - will not proceed otherwise.
    """

    # Approve FSM transition logic:

    def can_proceed(self, approval, check_conditions=True, **kwargs):
        """Return True if signoffs on this approval can proceed -- including that it's transition can proceed"""
        import django_fsm

        transition = self.bound_approve_transition(approval)
        fsm_can_proceed = (
            django_fsm.can_proceed(transition, check_conditions=check_conditions)
            if transition
            else True
        )
        return super().can_proceed(approval) and fsm_can_proceed

    def has_approval_transition_perm(self, approval, user, **kwargs):
        """Returns True iff model in state allows transition from given approval by given user"""
        import django_fsm

        transition = self.bound_approve_transition(approval)
        fsm_has_transition_perm = (
            django_fsm.has_transition_perm(transition, user) if transition else True
        )
        return (
            super().has_approval_transition_perm(approval, user, **kwargs)
            and fsm_has_transition_perm
        )

    def can_do_approve_transition(self, approval, user, **kwargs):
        """Return True iff all conditions are met for user to proceed with approval and make transition"""
        import django_fsm

        transition = self.bound_approve_transition(approval)
        fsm_can_proceed = (
            (
                django_fsm.can_proceed(transition)
                and django_fsm.has_transition_perm(transition, user)
            )
            if transition
            else True
        )
        return (
            super().can_do_approve_transition(approval, user, **kwargs)
            and fsm_can_proceed
        )

    # Revoke FSM transition logic:

    def is_revokable(self, approval, check_conditions=True, **kwargs):
        """Return True if the transition triggered by revoking approval (or approval name) can proceed"""
        import django_fsm

        transition = self.bound_revoke_transition(approval)
        fsm_can_proceed = (
            django_fsm.can_proceed(transition, check_conditions=check_conditions)
            if transition
            else False
        )
        return super().is_revokable(approval, **kwargs) and fsm_can_proceed

    def has_revoke_transition_perm(self, approval, user, **kwargs):
        """Returns True iff model in state allows transition for revoking given approval by given user"""
        import django_fsm

        transition = self.bound_revoke_transition(approval)
        fsm_has_transition_perm = (
            django_fsm.has_transition_perm(transition, user) if transition else False
        )
        return (
            super().has_revoke_transition_perm(approval, user, **kwargs)
            and fsm_has_transition_perm
        )

    def can_do_revoke_transition(self, approval, user, **kwargs):
        """Return True iff all conditions are met for user to proceed with revoke and make transition"""
        import django_fsm

        transition = self.bound_revoke_transition(approval)
        fsm_can_proceed = (
            (
                django_fsm.can_proceed(transition)
                and django_fsm.has_transition_perm(transition, user)
            )
            if transition
            else False
        )
        return (
            super().can_do_revoke_transition(approval, user, **kwargs)
            and fsm_can_proceed
        )


class ApprovalProcessDescriptor:
    """
    Descriptor provides a set of decorators to conveniently load a transition registry,
     and inject an ApprovalsProcess object that manages it in place of the descriptor attribute.
    """

    transition_registry_class = ApprovalTransitionRegistry
    approval_process_class = BasicApprovalProcess

    def __init__(
        self, *approval_sequence, transition_registry=None, approval_process_class=None
    ):
        """
        Optionally: define a linear approval sequence - default is sequenced by order approvals are registered
        Optionally: override the registry and process classes to use if the defaults are not suitable
        """
        self.registry = (
            transition_registry()
            if transition_registry
            else self.transition_registry_class()
        )
        self.approval_process_class = (
            approval_process_class or self.approval_process_class
        )
        # This allows one to explicitly set the approval sequence rather than relying on order of transition functions
        for a in approval_sequence:
            self.registry.add_approval(self._get_approval_id(a))

    @staticmethod
    def _get_approval_id(approval_descriptor_or_type):
        """ "approval_type" may be a string approval id, an actual ApprovalType, or an ApprovalDescriptor"""
        return (
            approval_descriptor_or_type
            if isinstance(approval_descriptor_or_type, str)
            else getattr(
                approval_descriptor_or_type,
                "approval_type",
                approval_descriptor_or_type,
            ).id
        )

    # Approve transition decorators
    # Typical usage (order matters!) (but don't actually do this... see convenience decorators below):
    #   @process.register_approve_transition(my_approval)
    #   @transition(my_state_variable, source=STATE1, target=STATE2, ...)  # optional FSM transition decorator
    #   @process.do_approval
    #   def approve_the_thing(self, approval, ...):
    #       ... other side-effects that should occur when the approval is approved
    #   ...
    #   if thing.process.can_do_approval_transition(approval_insance, user):
    #       save = thing.approve_the_thing(approval_instance)
    #       save()    # saves the approval and the thing in a DB transaction

    def register_approve_transition(self, approval_descriptor_or_type):
        """
        Return a decorator that simply adds the decorated function to the approval_process registry as the
        method to call to approve the given approval_type or ApprovalField descriptor
        """

        def register(transition_method):
            approval_type = self._get_approval_id(approval_descriptor_or_type)
            self.registry.add_approve_transition(approval_type, transition_method)
            return transition_method

        return register

    def do_approval(self, transition_method):
        """delegate"""
        return self.approval_process_class.do_approval(transition_method)

    def register_and_do_approval(self, approval_descriptor_or_type):
        """
        Convenience method to encapsulate register_approve_transition and do_approval
        Return a decorator that wraps fn with do_approval and registers the decorated fn
        See self.do_approval and self.register_approval_method
        Usage:
            @process.register_and_do_approval(my_approval)
            def approve_the_thing(self, approval, ...):
                ... other side-effects that should occur when the approval is approved
            ...
            thing.process.try_approve_transition(approval_instance, user)
        """
        register = self.register_approve_transition(approval_descriptor_or_type)

        def approval_transition_decorator(transition_method):
            """Decorate & register a transition_method with logic to complete approval"""
            return register(self.do_approval(transition_method))

        return approval_transition_decorator

    # Revoke transition decorators
    # Typical usage (order matters!) (but don't actually do this... see convenience decorators below):
    #   @process.register_revoke_transition(my_approval)
    #   @transition(my_state_variable, source=STATE2, target=STATE1, ...)  # optional FSM transition decorator
    #   @process.do_revoke
    #   def revoke_the_thing(self, approval, ...):
    #       ... other side-effects that should occur when the approval is revoked
    #   ...
    #   if thing.process.can_do_revoke_transition(approval_instance, user)
    #       revoke = thing.revoke_the_thing(approval_instance)
    #       revoke(request.user)    # revokes the approval and saves the thing in a DB transaction

    def register_revoke_transition(self, approval_descriptor_or_type):
        """
        Return a decorator that simply adds the decorated function to the approval_process registry as the
        method to call to revoke the given approval_type or ApprovalField descriptor
        """

        def register(transition_method):
            approval_type = self._get_approval_id(approval_descriptor_or_type)
            self.registry.add_revoke_transition(approval_type, transition_method)
            return transition_method

        return register

    def do_revoke(self, transition_method):
        """delegate"""
        return self.approval_process_class.do_revoke(transition_method)

    def register_and_do_revoke(self, approval_descriptor_or_type):
        """
        Convenience method to encapsulate register_revoke_transition and do_revoke
        Return a decorator that wraps fn with do_revoke and registers the decorated fn
        See self.do_revoke and self.register_revoke_method
        Usage:
            @process.register_and_do_revoke(my_approval)
            def revoke_the_thing(self, approval, ...):
                ... other side-effects that should occur when the approval is approved
            ...
            thing.process.try_revoke_transition(approval_instance, user)
        """
        register = self.register_revoke_transition(approval_descriptor_or_type)

        def revoke_transition_decorator(transition_method):
            """Decorate & register a transition_method w/ logic to revoke approval"""
            return register(self.do_revoke(transition_method))

        return revoke_transition_decorator

    def __set_name__(self, owner, name):
        """Grab the field name used by owning class to refer to this descriptor"""
        self.accessor_attr = name

    def __get__(self, instance, owner=None):
        """Return an ApprovalsProcess object to manage the transition registry for the given instance"""
        if not instance:
            return self
        else:
            approval_process = self.approval_process_class(
                process_model=instance, transition_registry=self.registry
            )
            setattr(instance, self.accessor_attr, approval_process)
            return approval_process


ApprovalsProcess = ApprovalProcessDescriptor  # A nicer name


class FsmApprovalProcessDescriptor(ApprovalProcessDescriptor):
    """Uses FSM Approval Actions to manage FSM state transitions"""

    approval_process_class = FsmApprovalProcess

    def approval_transition(
        self,
        approval_descriptor_or_type,
        field,
        source="*",
        target=None,
        on_error=None,
        conditions=None,
        permission=None,
        custom=None,
    ):
        """
        Convenience method to encapsulate registering and FSM transition that also completes an approval
        Return a decorator that wraps fn with both an FSM transition and do_approval, and registers the decorated fn
        All parameters except first positional are simply passed through to django_fsm.transition
        See django.fsm.transition, self.do_approval, and self.register_approval_method
        Usage:
            @process.approval_transition(my_approval, my_state_variable, source=STATE1, target=STATE2, ...)
            def approve_the_thing(self, approval, ...):
                ... other side-effects that should occur when the approval is approved
            ...
            thing.process.try_approve_transition(approval_instance, user)
        """
        import django_fsm

        register = self.register_approve_transition(approval_descriptor_or_type)
        fsm_decorator = django_fsm.transition(
            field,
            source,
            target,
            on_error=on_error,
            conditions=conditions or [],
            permission=permission,
            custom=custom or {},
        )

        def approval_transition_decorator(transition_method):
            """Decorate & register a transition_method with the fsm_decorator + logic to complete approval"""
            return register(fsm_decorator(self.do_approval(transition_method)))

        return approval_transition_decorator

    def revoke_transition(
        self,
        approval_descriptor_or_type,
        field,
        source="*",
        target=None,
        on_error=None,
        conditions=None,
        permission=None,
        custom=None,
    ):
        """
        Convenience method to encapsulate registering and FSM transition that also revokes an approval
        Return a decorator that wraps fn with both an FSM transition and do_revokde, and registers the decorated fn
        All parameters except first positional are simply passed through to django_fsm.transition
        See django.fsm.transition, self.do_revoke, and self.register_revoke_method
        Usage:
            @process.revoke_transition(my_approval, my_state_variable, source=STATE1, target=STATE2, ...)
            def revoke_the_thing(self, approval, ...):
                ... other side-effects that should occur when the approval is revoked
            ...
            thing.process.try_revoke_transition(approval_instance, user)
        """
        import django_fsm

        register = self.register_revoke_transition(approval_descriptor_or_type)
        fsm_decorator = django_fsm.transition(
            field,
            source,
            target,
            on_error=on_error,
            conditions=conditions or [],
            permission=permission,
            custom=custom or {},
        )

        def revoke_transition_decorator(transition_method):
            """Decorate & register a transition_method with the fsm_decorator + logic to complete approval"""
            return register(fsm_decorator(self.do_revoke(transition_method)))

        return revoke_transition_decorator


FsmApprovalsProcess = FsmApprovalProcessDescriptor  # A nicer name


def user_can_revoke_approval(approval_descriptor):
    """
    Return a callable suitable to pass as permission argument to `fsm.transition`

    Specifically intended for use within an `FsmApprovalsProcess` where a `RelatedApprovalDescriptor`,
    usually obtained from an `ApprovalField` is required to check the permission  for a `transition`

    :param approval_descriptor: a descriptor for an Approval that will be used to define the permission
        for an FSM transition defined in the same class.

    Usage:
    ```
        class MyProcess(models.Model):
            ...
            my_approval, my_approval_stamp = ApprovalField(.....)
            ...
            @fsm.transition(..., permission=user_can_revoke_approval(my_approval))
            def approve_it(self, approval):
                ...
    ```
    """

    def has_revoke_perm(instance, user):
        """Determine if the user has permission to revoke instance.approval"""
        approval = approval_descriptor.__get__(instance, type(instance))
        return approval.can_revoke(user)

    return has_revoke_perm


__all__ = [
    "ApprovalsProcess",
    "FsmApprovalsProcess",
    "user_can_revoke_approval",
]



"""
Django settings for django-signoffs tests.
"""
import os

ALLOWED_HOSTS = ("127.0.0.1", "localhost")

SECRET_KEY = "django-insecure"
DEBUG = True

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

INSTALLED_APPS = (
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.auth",
    "django.contrib.admin",
    "django.contrib.messages",
    "django_fsm",
    "django.contrib.staticfiles",
    # project apps
    "signoffs",
    "signoffs.contrib.signets",
    "signoffs.contrib.approvals",
    "tests.test_app",
    "demo",
    "demo.article",
    "demo.assignments",
    "demo.registration",
)

MIDDLEWARE = [
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

ROOT_URLCONF = "tests.test_app.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
        "TEST": {
            "NAME": None,  # use in-memory test DB
            "MIGRATE": False,  # Django 3.1+ -- disable migrations, create test DB schema directly from models.
        },
    }
}


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.2/howto/static-files/

STATIC_URL = "/static/"

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"



"""
Tests for signoff Types registries
"""
from django.test import SimpleTestCase

from signoffs.registry import signoffs
from tests.test_app import models


class TestSignoffsRegistry(SimpleTestCase):
    def test_signoff_registered(self):
        s = signoffs.get("test_app.agree")
        self.assertEqual(s().signet_model, models.Signet)
        s = signoffs.get("test_app.accept")
        self.assertEqual(s().signet_model, models.ReportSignet)

    def test_signoff_type(self):
        o = models.Signet(signoff_id="test_app.agree")
        self.assertEqual(o.signoff_type, signoffs.get("test_app.agree"))



"""
    Objects that know how to render other Signoffs / Signets into HTML
"""
from django.template.loader import render_to_string

from signoffs.core import utils

from . import helpers


class SignoffInstanceRenderer:
    """Renderer for a Signoff instance"""

    signet_template = "signoffs/signets/signet.html"
    signoff_form_template = "signoffs/signets/signoff_form.html"

    # default template context values - can be overridden with context variables or template tag kwargs,
    #    or override defaults by subclassing, or by passing overrides dict to init.
    signet_context = dict(
        show_revoke=True,
        timestamp_label="Date",
        with_signature_line=False,
        signature_line_label="Signature",
    )

    form_context = dict(
        show_form=True,
        help_text="Check box and Sign to add your consent.",
        submit_label="Sign",
        signoff_form=None,
        csrf_token=None,
    )

    pass_thru_context = (
        "request",
        "csrf_token",
        "request_user",
    )  # variables passed through to template from parent context

    def __init__(
        self,
        signoff_instance,
        signet_template=None,
        signet_context=None,
        signoff_form_template=None,
        form_context=None,
    ):
        """A renderer instance for given signoff_type, optionally override class templates"""
        self.signoff = signoff_instance
        self.signet_template = signet_template or self.signet_template
        # Force request into context so it is available from context being rendered in
        self.signet_context = {
            **self.signet_context,
            **(signet_context or {}),
            **{v: None for v in self.pass_thru_context},
        }
        # + for Signoff Form:
        self.signoff_form_template = signoff_form_template or self.signoff_form_template
        self.form_context = {
            **self.form_context,
            **(form_context or {}),
            **{v: None for v in self.pass_thru_context},
        }

    def __call__(self, request_user=None, context=None, **kwargs):
        """Return a string containing a rendered version of this signoff, optionally tailored for requesting user."""
        return (
            self.signet(request_user, context, **kwargs)
            if self.signoff.is_signed()
            else self.form(request_user, context, **kwargs)
        )

    def signet(self, request_user=None, context=None, **kwargs):
        """Return a string containing the rendered Signet for given user, if it is signed, empty string otherwise"""
        context = context or {}
        request_user = helpers.resolve_request_user(request_user, context, **kwargs)
        show_revoke = kwargs.pop(
            "show_revoke", self.signet_context.get("show_revoke", True)
        )
        signet_template = kwargs.pop("signet_template", self.signet_template)
        return (
            render_to_string(
                signet_template,
                helpers.resolve_dicts(
                    defaults=self.signet_context,
                    overrides=context,
                    signoff=self.signoff,
                    request_user=request_user,
                    is_revokable=show_revoke
                    and request_user
                    and self.signoff.can_revoke(request_user),
                    **kwargs,
                ),
            )
            if self.signoff.is_signed()
            else ""
        )

    def form(self, request_user=None, context=None, **kwargs):
        """Return a string containing the rendered Signet Form, if it can be signed, empty string otherwise"""
        context = context or {}
        request_user = helpers.resolve_request_user(request_user, context, **kwargs)
        show_form = kwargs.pop("show_form", self.form_context.get("show_form", True))
        form_template = kwargs.pop("signoff_form_template", self.signoff_form_template)
        return (
            render_to_string(
                form_template,
                self.resolve_form_context(
                    overrides=context,
                    signoff=self.signoff,
                    request_user=request_user,
                    is_signable=show_form
                    and request_user
                    and self.signoff.can_sign(request_user),
                    **kwargs,
                ),
            )
            if request_user and self.signoff.can_sign(request_user)
            else ""
        )

    def resolve_form_context(self, overrides, **kwargs):
        """return single context dictionary suitable for rendering signet form"""
        form_context = self.form_context.copy()
        default_label = form_context.pop("label", self.signoff.label)
        default_help_text = form_context.pop("help_text", None)
        form_class = self.signoff.forms.get_signoff_form_class(
            signoff_field_kwargs=dict(
                label=kwargs.get("label", default_label),
                help_text=kwargs.get("help_text", default_help_text),
            )
        )
        form_context.update(dict(signoff_form=form_class(instance=self.signoff)))
        form_context.update(
            helpers.filter_dict(overrides, self.form_context.keys())
        )  # overrides from context
        form_context.update(kwargs)  # kwargs override all
        return form_context


class SignoffRenderer(utils.service(SignoffInstanceRenderer)):
    """
    A descriptor class that "injects" a `SignoffInstanceRenderer` instance into a Signoff instance.

    To inject custom rendering services:
      - provide a custom service_class:  `render=SignoffRenderer(service_class=MyInstanceRenderer)`
      - OR specialize class attributes:
        `MyRenderer = utils.service(SignoffInstanceRenderer, signet_template='my.tmpl.html')`
      - OR both... `MyRenderer = utils.service(MyInstanceRenderer)`
    """



"""
    Objects that know how to render an Approval Process into HTML
"""
from django.template.loader import render_to_string

from signoffs.core import utils

from . import helpers


class ApprovalProcessInstanceRenderer:
    """Renderer for an ApprovalProcess instance"""

    process_template = "signoffs/process/approval_process.html"

    # default template context values - can be overridden with context variables or template tag kwargs,
    #    or override defaults by subclassing, or by passing overrides dict to init.
    approval_context = dict(
        show_revoke=True,
        show_status_msg=True,
        render_signoff_forms=True,
    )

    pass_thru_context = (
        "request",
        "csrf_token",
        "request_user",
    )  # variables passed through to template from parent context

    def __init__(self, approval_process, process_template=None, approval_context=None):
        """A renderer instance for given approval_type, optionally override class templates"""
        self.process = approval_process
        self.process_template = process_template or self.process_template
        # Force request into context so it is available from context being rendered in
        self.approval_context = {
            **self.approval_context,
            **(approval_context or {}),
            **{v: None for v in self.pass_thru_context},
        }

    def __call__(self, request_user=None, context=None, **kwargs):
        """Return a string containing a rendered version of this approval, optionally tailored for requesting user."""
        context = context or {}
        request_user = helpers.resolve_request_user(request_user, context, **kwargs)
        return render_to_string(
            self.process_template,
            helpers.resolve_dicts(
                defaults=self.approval_context,
                overrides=context,
                process=self.process,
                request_user=request_user,
                **kwargs,
            ),
        )


class ApprovalProcessRenderer(utils.service(ApprovalProcessInstanceRenderer)):
    """
    A descriptor class that "injects" a `ApprovalProcessInstanceRenderer` instance into a ApprovalProceess instance.

    To inject custom rendering services:
      - provide a custom service_class:  `render=ApprovalProcessRenderer(service_class=MyInstanceRenderer)`
      - OR specialize class attributes:
        `MyRenderer = utils.service(ApprovalProcessInstanceRenderer, process_template='my.tmpl.html')`
      - OR both... `MyRenderer = utils.service(MyInstanceRenderer)`
    """



"""
    Proxy for Signoff Signing Order to simplify import statements and hide core package structure from client code.

    isort:skip_file
"""
from signoffs.core.signing_order.signing_order import (
    SigningOrderStrategyProtocol,
    SigningOrder,
)
from signoffs.core.signing_order.signoff_pattern import (
    AnyOneOf,
    AtLeastN,
    ExactlyN,
    ExactlyOne,
    InParallel,
    InSeries,
    OneOrMore,
    Optional,
)



"""
App-independent tests for Signoff model descriptors - no app logic
"""
from django.contrib.auth.models import AnonymousUser
from django.test import TestCase

from signoffs.core.forms import AbstractSignoffForm

from . import fixtures
from .models import InvalidModel, LeaveRequest, Signet


class SimpleSignoffRelationTests(TestCase):
    def test_signofffield_relations(self):
        lr = LeaveRequest()
        self.assertEqual(
            LeaveRequest.employee_signoff.id, LeaveRequest.employee_signoff_type.id
        )
        self.assertEqual(type(lr.employee_signoff).id, LeaveRequest.employee_signoff.id)
        self.assertEqual(
            type(lr.employee_signoff.signet),
            LeaveRequest.employee_signoff.get_signetModel(),
        )
        self.assertEqual(lr.employee_signet, None)

    def test_signofffield(self):
        lr = LeaveRequest()
        # OneToOne forward relation
        self.assertTrue(
            isinstance(lr.employee_signoff, LeaveRequest.employee_signoff_type)
        )
        self.assertEqual(lr.employee_signoff.signet_model, Signet)
        self.assertFalse(lr.employee_signoff.is_signed())

    def test_signoffset(self):
        lr = LeaveRequest.objects.create()
        # ManyToOne reverse relation
        self.assertEqual(lr.hr_signoffs.count(), 0)

    def test_signoffset_integrity_checks(self):
        m = InvalidModel()
        with self.assertRaises(ValueError):
            m.invalid_signet.count()
        with self.assertRaises(ValueError):
            m.invalid_relation.count()


class SignoffRelationTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        u1 = fixtures.get_user()
        u2 = fixtures.get_user()
        u3 = fixtures.get_user()
        lr = LeaveRequest.objects.create()
        lr.employee_signoff.sign(u1)
        cls.hr_signoffs = (
            lr.hr_signoffs.create(user=u1),
            lr.hr_signoffs.create(user=u2),
        )
        lr.mngmt_signoffs.create(user=u3),
        cls.u1, cls.u2, cls.u3 = u1, u2, u3
        cls.lr = lr

    def test_signofffield(self):
        with self.assertNumQueries(1):
            lr = LeaveRequest.objects.select_related("employee_signet").get(
                pk=self.lr.pk
            )
            # OneToOne forward relation
            self.assertTrue(lr.employee_signoff.is_signed())

    def test_signoffset_manager(self):
        with self.assertNumQueries(2):
            lr = LeaveRequest.objects.prefetch_related("signatories").get(pk=self.lr.pk)
            # ManyToOne reverse relation
            self.assertTrue(lr.hr_signoffs.exists())
            self.assertEqual(lr.hr_signoffs.count(), 2)
            self.assertTrue(all(s.is_signed() for s in lr.hr_signoffs.all()))

    def test_signoffset_queries(self):
        with self.assertNumQueries(2):
            lr = LeaveRequest.objects.prefetch_related("signatories").get(pk=self.lr.pk)
            # Earliest / Latest
            self.assertEqual(lr.hr_signoffs.earliest(), self.hr_signoffs[0])
            self.assertEqual(lr.hr_signoffs.latest(), self.hr_signoffs[-1])

    def test_signoffset_signatories(self):
        with self.assertNumQueries(3):
            lr = LeaveRequest.objects.prefetch_related("signatories__user").get(
                pk=self.lr.pk
            )
            self.assertTrue(lr.hr_signoffs.can_sign(self.u3))
            self.assertFalse(lr.hr_signoffs.can_sign(AnonymousUser()))
            self.assertTrue(lr.hr_signoffs.has_signed(self.u1))
            self.assertFalse(lr.hr_signoffs.has_signed(self.u3))

    def test_signoffset_form(self):
        lr = LeaveRequest.objects.prefetch_related("signatories").get(pk=self.lr.pk)
        form = lr.hr_signoffs.forms.get_signoff_form_class()
        self.assertTrue(issubclass(form, AbstractSignoffForm))
        self.assertEqual(form()["signoff_id"].initial, LeaveRequest.hr_signoff_type.id)

    def test_signoffset_revoked(self):
        u = fixtures.get_user()
        lr = LeaveRequest.objects.prefetch_related("signatories").get(pk=self.lr.pk)
        n_revokes = 3
        for _ in range(n_revokes):
            so = lr.hr_signoffs.create(user=u)
            so.revoke(user=u, reason="just because")
        with self.assertNumQueries(2):
            lr = LeaveRequest.objects.get(pk=self.lr.pk)
            revoked = lr.hr_signoffs.revoked()
            self.assertEqual(len(revoked), n_revokes)
            self.assertTrue(all(s.revoked.user == u for s in revoked))


# DEV Testing only - delete me


# class DevTests(TestCase):
#
#     def dont_test_it(self):
#         lr = LeaveRequest.objects.create()
#         print([
#             (f.related_model, f.model) for f in LeaveRequest.hr_signoff_type.get_signetModel()._meta.fields
#                 if isinstance(f, models.ForeignKey)
#                    and not issubclass(f.related_model, AbstractSignet)
#                    and f.name != 'user'
#         ])
#
#         self.assertTrue(False)



"""
    contrib.approvals adds basic concrete ApprovalStamp-related models, signoffs, and approvals.
"""
from django.apps import AppConfig


class ContribApprovalsConfig(AppConfig):
    name = "signoffs.contrib.approvals"
    label = "signoffs_approvals"
    default = True
    default_auto_field = "django.db.models.BigAutoField"



""" Helpers: resolve 3 potential sources for context: defaults, context object, kwargs """


# Helper methods:


def resolve_request_user(request_user, context, **kwargs):
    """return user object either from request user or context.request.user or None"""
    # Only need the request.user, so don't require a request object, but often convenient to use one  ** sigh **
    request_user = request_user or kwargs.get(
        "request_user", context.get("request_user", None)
    )
    request = kwargs.get("request", context.get("request", None))
    return request_user or (request.user if request else None)


def filter_dict(d, keys):
    """Return copy of d filtered by keys"""
    d = d or {}
    return {k: d.get(k) for k in keys if k in d}


def resolve_dicts(defaults, overrides, **kwargs):
    """return a single dictionary, overriding matching keys from overrides and updating with kwargs"""
    context = defaults.copy()  # defaults: lowest precedence
    context.update(filter_dict(overrides, defaults.keys()))  # overrides
    context.update(kwargs)  # kwargs take precedence
    return context



"""
    Basic concrete implementation for Signet models
"""

from signoffs.models import AbstractRevokedSignet, AbstractSignet


class Signet(AbstractSignet):
    """
    A concrete persistence layer for basic Signoffs with no relations.

    Suitable for out-of-the-box use with `signoffs.models.SignoffField`
    """

    pass


class RevokedSignet(AbstractRevokedSignet):
    """
    A concrete persistence layer for revoked Signoffs.

    May be declared on Signoff Types to provide persistence and tracking of revoked signoffs.
    """

    pass



"""
Test App - tests for shortcuts
"""
from django.test import TestCase

from signoffs import shortcuts
from signoffs.core.tests import fixtures
from tests.test_app import models, signoffs


class ShortcutTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user = fixtures.get_user()
        cls.signoff = signoffs.agree_signoff.create(cls.user)
        cls.approval = models.SimpleApproval.create()

    def test_get_signet_or_404(self):
        signet = shortcuts.get_signet_or_404(self.signoff.id, self.signoff.signet.pk)
        self.assertEqual(signet, self.signoff.signet)

    def test_get_signoff_or_404(self):
        so = shortcuts.get_signoff_or_404(self.signoff.id, self.signoff.signet.pk)
        self.assertEqual(so, self.signoff)

    def test_get_approval_stamp_or_404(self):
        stamp = shortcuts.get_approval_stamp_or_404(
            self.approval.id, self.approval.stamp.pk
        )
        self.assertEqual(stamp, self.approval.stamp)

    def test_get_approval_or_404(self):
        approval = shortcuts.get_approval_or_404(
            self.approval.id, self.approval.stamp.pk
        )
        self.assertEqual(approval, self.approval)
